# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from dataclasses import (
    asdict,
    dataclass,
)
from typing import (
    Any,
    Callable,
    Optional,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

from boulderopal._core import closed_loop_step_workflow
from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
)
from boulderopal.closed_loop._optimizers import (
    ALLOWED_OPTIMIZERS,
    ClosedLoopOptimizer,
    Cmaes,
    GaussianProcess,
    NeuralNetwork,
    SimulatedAnnealing,
)


@dataclass
class Results:
    """
    Results from evaluating the closed-loop optimization cost function.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters at which the cost function was evaluated
        as a 2D array of shape ``(test_point_count, parameter_count)``.
    costs : np.ndarray
        The evaluated costs from the cost function
        as a 1D array of shape ``(test_point_count,)``.
    cost_uncertainties : np.ndarray or None, optional
        The uncertainties associated with the costs.
        If provided, must have the same shape as costs.
        Defaults to None, in which case there is no uncertainty associated to the costs.

    See Also
    --------
    boulderopal.closed_loop.step : Perform a single step in a closed-loop optimization.
    """

    parameters: np.ndarray
    costs: np.ndarray
    cost_uncertainties: Optional[np.ndarray] = None

    def __post_init__(self) -> None:
        self.parameters = ArrayDType.REAL(self.parameters, "parameters", ndim=2)
        self.costs = ArrayDType.REAL(self.costs, "costs", ndim=1)
        check_argument(
            self.costs.shape == (self.parameters.shape[0],),
            "The costs must be a 1D array with one item per measurement.",
            {"costs": self.costs, "parameters": self.parameters},
            extras={"parameters.shape[0]": self.parameters.shape[0]},
        )

        if self.cost_uncertainties is not None:
            self.cost_uncertainties = ArrayDType.REAL(
                self.cost_uncertainties, "cost_uncertainties", ndim=1
            )
            check_argument(
                self.cost_uncertainties.shape == self.costs.shape,
                "The costs uncertainties must be an array with the same shape as the costs.",
                {"cost_uncertainties": self.cost_uncertainties, "costs": self.costs},
            )


def step(
    optimizer: ClosedLoopOptimizer | str,
    results: Optional[Results] = None,
    test_point_count: Optional[int] = None,
) -> dict:
    """
    Perform a single step in a closed-loop optimization.

    Parameters
    ----------
    optimizer : ClosedLoopOptimizer or str
        The optimizer to be used in the minimization of the cost function, or an optimizer state.
        If this is the first optimization step pass an instance of a closed-loop optimizer class.
        In subsequent steps, pass the optimizer state as returned by the previous step.
    results : Results or None, optional
        Results from evaluating the cost function. You can omit this parameter if you want to
        generate more test points but have not yet evaluated the cost function.
        Defaults to None.
        Note that some optimizers might have extra requirements of results needed for the first
        optimization step. You can check the documentation of the individual optimizers for details.
    test_point_count : int or None, optional
        Requested number of test points to be generated from the optimizer at this step.
        This is a hint to the optimizer, the number of returned test points might be different.
        You should set this value based on the number of test points you can efficiently calculate
        as a batch or in parallel. Typically, optimizers will return at least this many points,
        but they might return more (for example if a certain number of points is required in order
        to move the algorithm to the next state) or, occasionally, fewer (for example if moving the
        algorithm to the next state requires the evaluation of a specific point and nothing more).

    Returns
    -------
    dict
        A dictionary containing the optimization step result, with the following keys:

        ``test_points``
            New test points at which the cost function should be evaluated
            for the next optimization step.
        ``state``
            The optimizer state, to be provided in the next step.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    boulderopal.closed_loop.optimize :
        Run a closed-loop optimization to find a minimum of the given cost function.
    boulderopal.execute_graph : Evaluate generic functions.
    boulderopal.run_gradient_free_optimization :
        Perform model-based optimization without using gradient values.
    boulderopal.run_optimization :
        Perform gradient-based deterministic optimization of generic real-valued functions.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.
    """
    check_argument(
        isinstance(optimizer, (*ALLOWED_OPTIMIZERS, str)),
        "You must provide either an optimizer or an optimizer state.",
        {"optimizer": optimizer},
    )

    if isinstance(optimizer, ClosedLoopOptimizer):
        closed_loop_optimizer: dict[str, Any] | str = asdict(optimizer)
    else:
        closed_loop_optimizer = optimizer

    test_point_count = nullable(
        ScalarDType.INT, test_point_count, "test_point_count", min_=0
    )

    return closed_loop_step_workflow(
        optimizer=closed_loop_optimizer,
        results=None if results is None else asdict(results),
        test_point_count=test_point_count,
    )


def _generate_random_inital_points(optimizer: ClosedLoopOptimizer) -> np.ndarray:
    """
    Generate a random array of initial points depending on the chosen optimizer.
    """
    parameter_count = optimizer.bounds.values.shape[0]

    if isinstance(optimizer, Cmaes):
        if optimizer.population_size is not None:
            point_count = optimizer.population_size
        else:
            point_count = int(4 + np.floor(3 * np.log(parameter_count)))

    elif isinstance(optimizer, (GaussianProcess, NeuralNetwork, SimulatedAnnealing)):
        point_count = 2 * parameter_count

    else:
        # This should never be raised as we validate the optimizer in the main function.
        raise QctrlArgumentsValueError(
            "The type of the optimizer is not valid.", {"optimizer": optimizer}
        )

    assert hasattr(optimizer, "seed")  # Make mypy happy.
    rng = np.random.default_rng(seed=optimizer.seed)
    bounds_ = optimizer.bounds.values
    return rng.uniform(
        low=bounds_[:, 0], high=bounds_[:, 1], size=(point_count, len(bounds_))
    )


def optimize(
    cost_function: Callable,
    optimizer: ClosedLoopOptimizer | str,
    initial_parameters: Optional[np.ndarray] = None,
    target_cost: Optional[float] = None,
    max_iteration_count: int = 100,
    callback: Optional[Callable] = None,
    verbose: bool = True,
) -> dict:
    """
    Run a closed-loop optimization to find a minimum of the given cost function.

    This is an iterative process, where the optimizer generates and tests a set of points.
    After several iterations the distribution of generated test points should converge
    to low values of the cost function. You can use this approach when your system is too
    complicated to model.

    The provided cost function must take a 2D array of shape ``(test_point_count, parameter_count)``
    as input and return a 1D array of costs of length `test_point_count`.
    Alternatively, it can return a 2D array of shape ``(2, test_point_count)``, where the first
    row represents the costs and the second row represents their associated uncertainties.

    For best results, you should provide a set of `initial_parameters` to start the optimization.
    The performance and convergence of the optimizer might change depending on these values.
    If you don't pass `initial_parameters`, randomly sampled values inside the bounds are used.
    The number of initial values is set to the population size for CMA-ES and to
    ``2 * parameter_count`` for other optimizers.

    Parameters
    ----------
    cost_function : Callable
        The cost function to minimize, as a callable that takes a 2D NumPy array of parameters and
        returns either a 1D NumPy array of costs, or a 2D NumPy arrays of costs and uncertainties.
        The cost function should always return the same type of output
        (either always return uncertainties or never return them).
    optimizer : ClosedLoopOptimizer or str
        The optimizer to be used in the minimization of the cost function, or an optimizer state.
        If this is the first optimization step, pass an instance of a closed-loop optimizer class.
        If you want to resume an optimization, pass the optimizer state as of the last step.
    initial_parameters : np.ndarray or None, optional
        The initial values of the parameters to use in the optimization,
        as a 2D NumPy array of shape ``(test_point_count, parameter_count)``.
        If not passed, random values uniformly sampled inside the optimizer bounds are used.
        If you provide an optimizer state, you must provide an array of initial parameters.
    target_cost : float or None, optional
        The target cost.
        If passed, the optimization will halt if the best cost is below the given value.
    max_iteration_count : int, optional
        The maximum number of iterations.
        Defaults to 100.
    callback : Callable or None, optional
        A function that takes in the current set of parameters, a 2D NumPy array of shape
        ``(test_point_count, parameter_count)``, and returns a bool.
        The function is evaluated once during each iteration with the
        current parameters. If it returns True, the optimization is halted.
    verbose : bool, optional
        Whether to print out information about the optimization cycle.
        Defaults to True.

    Returns
    -------
    dict
        A dictionary containing the optimization result, with the following keys:

        ``cost``
            The lowest cost found during the optimization.
        ``parameters``
            The optimal parameters associated to the lowest cost.
        ``cost_history``
            The history of best cost values up to each optimization step.
        ``step``
            A dictionary containing the information about the last optimization step,
            to be used to resume the optimization.
            It contains the optimizer ``state`` and the ``test_points`` at which
            the cost function should be evaluated next.

    See Also
    --------
    boulderopal.closed_loop.step :
        Perform a single step in a closed-loop optimization.
    boulderopal.execute_graph : Evaluate generic functions.
    boulderopal.run_gradient_free_optimization :
        Perform model-based optimization without using gradient values.
    boulderopal.run_optimization :
        Perform gradient-based deterministic optimization of generic real-valued functions.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.

    Notes
    -----
    At each iteration, the cost function will be called with the same number of data points as
    initially provided in `initial_parameters`. However, in some situations the optimizer might
    request more points (for example, if a certain number of points is required in order to move
    the algorithm to the next state) or, occasionally, fewer (for example, if moving the algorithm
    to the next state requires the evaluation of a specific point and nothing more).

    If the optimization loop is halted via a `KeyboardInterrupt` then the function returns
    the best results obtained in the optimization thus far.
    """

    def verbose_print(message: str) -> None:
        if verbose:
            print(message)

    initial_parameters = nullable(
        ArrayDType.REAL, initial_parameters, "initial_parameters", ndim=2
    )

    check_argument(
        isinstance(optimizer, (*ALLOWED_OPTIMIZERS, str)),
        "You must provide either an optimizer or an optimizer state.",
        {"optimizer": optimizer},
    )

    if isinstance(optimizer, ClosedLoopOptimizer):
        if initial_parameters is not None:
            check_argument(
                np.all(
                    (initial_parameters >= optimizer.bounds.values[:, 0])
                    & (initial_parameters <= optimizer.bounds.values[:, 1])
                ),
                "The initial parameters must be within the optimizer bounds.",
                {"initial_parameters": initial_parameters},
                extras={"optimizer.bounds.values": optimizer.bounds.values},
            )
        else:
            initial_parameters = _generate_random_inital_points(optimizer)
        closed_loop_optimizer: dict | str = asdict(optimizer)
        verbose_print(
            "Running closed loop optimization\n"
            "-----------------------------------------------\n"
            f"  Optimizer             : {optimizer.method_name}\n"
            f"  Number of test points : {initial_parameters.shape[0]}\n"
            f"  Number of parameters  : {initial_parameters.shape[1]}\n"
            "-----------------------------------------------\n"
        )
    else:
        check_argument(
            initial_parameters is not None,
            "If you provide an optimizer state, you must provide initial parameters.",
            {"initial_parameters": initial_parameters, "optimizer": optimizer},
        )
        assert isinstance(initial_parameters, np.ndarray)  # Make mypy happy.

        closed_loop_optimizer = optimizer
        verbose_print(
            "Resuming closed loop optimization\n"
            "-----------------------------------------------\n"
            f"  Number of test points : {initial_parameters.shape[0]}\n"
            f"  Number of parameters  : {initial_parameters.shape[1]}\n"
            "-----------------------------------------------\n"
        )

    max_iteration_count = ScalarDType.INT(
        max_iteration_count, "max_iteration_count", min_=0
    )

    test_parameters = initial_parameters
    test_point_count = initial_parameters.shape[0]

    # Obtain initial costs.
    verbose_print("Calling cost function…")
    cost_return = cost_function(test_parameters)

    check_argument(
        isinstance(cost_return, np.ndarray) and cost_return.ndim in (1, 2),
        "The cost function must return a 1D or 2D array.",
        {"initial_parameters": initial_parameters},
        extras={"cost_function(test_parameters)": cost_return},
    )

    if cost_return.ndim == 2:
        check_argument(
            cost_return.shape == (2, len(initial_parameters)),
            "If the cost function returns a 2D array, it must be of shape "
            "``(2, test_point_count)``.",
            {"initial_parameters": initial_parameters},
            extras={"cost_function(test_parameters)": cost_return},
        )
        uncertainties_returned = True
        costs, cost_uncertainties = cost_return

    else:
        check_argument(
            cost_return.shape == (len(initial_parameters),),
            "If the cost function returns a 1D array, it must be of length `test_point_count`.",
            {"initial_parameters": initial_parameters},
            extras={"cost_function(test_parameters)": cost_return},
        )

        uncertainties_returned = False
        costs = cost_return
        cost_uncertainties = None

    best_cost_overall, best_parameters_overall = min(
        zip(costs, test_parameters), key=lambda params: params[0]
    )
    verbose_print(f"  Initial best cost: {best_cost_overall:.3e}")

    # Store the cost history.
    best_cost_history = [best_cost_overall]

    # Run the optimization loop until a halting condition is met.
    try:
        for iteration_count in range(max_iteration_count):
            results = Results(
                parameters=test_parameters,
                costs=costs,
                cost_uncertainties=cost_uncertainties,
            )

            # Call the automated closed-loop optimizer and obtain the next set of test points.
            verbose_print("\nRunning optimizer…")
            step_result = closed_loop_step_workflow(
                optimizer=closed_loop_optimizer,
                results=asdict(results),
                test_point_count=test_point_count,
            )

            # Retrieve the data returned by the closed-loop optimizer.
            closed_loop_optimizer = step_result["state"]
            test_parameters = step_result["test_points"]

            # Obtain costs.
            verbose_print("Calling cost function…")
            if uncertainties_returned:
                costs, cost_uncertainties = cost_function(test_parameters)
            else:
                costs = cost_function(test_parameters)

            # Record the best results after this iteration.
            best_cost, best_parameters = min(
                zip(costs, test_parameters), key=lambda params: params[0]
            )

            # Compare last best results with best result overall.
            if best_cost < best_cost_overall:
                best_cost_overall = best_cost
                best_parameters_overall = best_parameters

            # Print the current best cost.
            verbose_print(
                f"  Best cost after {iteration_count+1} iterations: {best_cost_overall:.3e}"
            )

            # Store the current best cost.
            best_cost_history.append(best_cost_overall)

            if callback is not None:
                if callback(test_parameters):
                    verbose_print(
                        "\nCallback condition satisfied. Stopping the optimization."
                    )
                    break

            # Check if desired threshold has been achieved.
            if target_cost is not None:
                if best_cost_overall < target_cost:
                    verbose_print("\nTarget cost reached. Stopping the optimization.")
                    break
        else:
            verbose_print(
                "\nMaximum iteration count reached. Stopping the optimization."
            )

    except KeyboardInterrupt:
        # Exit the loop and return the current best results if a KeyboardInterrupt is caught.
        print("\nCalculation interrupted. Stopping the optimization.")

    return {
        "cost": best_cost_overall,
        "parameters": best_parameters_overall,
        "cost_history": best_cost_history,
        "step": step_result,
    }



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from numbers import Number
from typing import (
    Tuple,
    Union,
)

import forge
import numpy as np
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_numeric_numpy_array,
)

from .documentation import Category
from .node_data import (
    Pwc,
    Stf,
    Tensor,
)
from .types import (
    NumericOrFunction,
    TensorLikeOrFunction,
)
from .utils import (
    validate_batch_and_value_shapes,
    validate_shape,
)


def check_operand_type(x, name, can_be_number):
    """
    Check the operand is a number, a NumPy array, a Pwc, an Stf, or a Tensor.

    The check also passes if the argument can possibly be a QuTiP operator (has
    a `full` argument).

    Parameters
    ----------
    x : Any
        The operand to be checked.
    name : str
        Name of the operand.
    can_be_number : bool
        Whether the operand can be number.
    """
    if can_be_number:
        types = (Pwc, Stf, Tensor, np.ndarray, Number)
        message = (
            f"The variable {name} must be a number, a NumPy array, a Pwc, an Stf, "
            "or a Tensor."
        )
    else:
        types = (Pwc, Stf, Tensor, np.ndarray)
        message = (
            f"The variable {name} must be a NumPy array, a Pwc, an Stf, or a Tensor."
        )
    check_argument(
        isinstance(x, types) or hasattr(x, "full"),
        message,
        {name: x},
        extras={f"type({name})": type(x)},
    )


def _create_flexible_unary_node_data(
    _operation, x, name, value_shape_changer=None, can_be_number=True
):
    """
    Common implementation of `create_node_data` for nodes acting on Tensors, Pwcs, and Stfs
    implementing unary functions.

    Parameters
    ----------
    _operation : Operation
        The operation to implement.
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object on which the operation acts.
    name : str
        The name of the node.
    value_shape_changer : Callable[[tuple], tuple], optional
        Callable that transforms the original shape of the object into the shape after the operation
        is applied. Defaults to an identity operation, that is to say, to not change the shape.
    can_be_number : bool
        Whether the input variable can be numeric scalar. Defaults to True.

    Returns
    -------
    Tensor or Pwc or Stf
        The operation acting on the object.
    """
    check_operand_type(x, "x", can_be_number)

    # By default don't change shapes.
    def get_value_shape(shape):
        if value_shape_changer is None:
            return shape
        return value_shape_changer(shape)

    if isinstance(x, Stf):
        check_argument(
            name is None, "You can't assign a name to an Stf node.", {"name": name}
        )
        batch_shape, value_shape = validate_batch_and_value_shapes(x, "x")
        return Stf(
            _operation,
            value_shape=get_value_shape(value_shape),
            batch_shape=batch_shape,
        )

    if isinstance(x, Pwc):
        batch_shape, value_shape = validate_batch_and_value_shapes(x, "x")
        return Pwc(
            _operation,
            value_shape=get_value_shape(value_shape),
            durations=x.durations,
            batch_shape=batch_shape,
        )

    check_numeric_numpy_array(x, "x")
    shape = validate_shape(x, "x")
    return Tensor(_operation, shape=get_value_shape(shape))


class Sqrt(Node):
    r"""
    Calculate the element-wise square root of an object. This can be a number, an array, a tensor,
    or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose square root you want to calculate, :math:`x`. For numbers, arrays, and
        tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise square root, :math:`\sqrt{x}`, of the values or function you provided.
        The returned object is of the same kind as the one you provided, except if you provide a
        number or an np.ndarray in which case it's a Tensor.
    """

    name = "sqrt"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Sin(Node):
    r"""
    Calculate the element-wise sine of an object. This can be a number, an array, a tensor,
    or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose sine you want to calculate, :math:`x`. For numbers, arrays, and
        tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise sine, :math:`\sin{x}`, of the values or function you provided.
        The returned object is of the same kind as the one you provided, except if you provide a
        number or an np.ndarray in which case it's a Tensor.
    """

    name = "sin"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Cos(Node):
    r"""
    Calculate the element-wise cosine of an object. This can be a number, an array, a tensor,
    or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose cosine you want to calculate, :math:`x`. For numbers, arrays, and
        tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise cosine, :math:`\cos{x}`, of the values or function you provided.
        The returned object is of the same kind as the one you provided, except if you provide a
        number or an np.ndarray in which case it's a Tensor.
    """

    name = "cos"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Tan(Node):
    r"""
    Calculate the element-wise tangent of an object. This can be a number, an array, a tensor,
    or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose tangent you want to calculate, :math:`x`. For numbers, arrays, and
        tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise tangent, :math:`\tan{x}`, of the values or function you provided.
        The returned object is of the same kind as the one you provided, except if you provide a
        number or an np.ndarray in which case it's a Tensor.
    """

    name = "tan"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Sinh(Node):
    r"""
    Calculate the element-wise hyperbolic sine of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose hyperbolic sine you want to calculate, :math:`x`. For numbers, arrays, and
        tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise hyperbolic sine, :math:`\sinh{x}`, of the values or function you provided.
        The returned object is of the same kind as the one you provided, except if you provide a
        number or an np.ndarray in which case it's a Tensor.
    """

    name = "sinh"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Cosh(Node):
    r"""
    Calculate the element-wise hyperbolic cosine of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose hyperbolic cosine you want to calculate, :math:`x`. For numbers, arrays,
        and tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise hyperbolic cosine, :math:`\cosh{x}`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "cosh"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Tanh(Node):
    r"""
    Calculate the element-wise hyperbolic tangent of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose hyperbolic tangent you want to calculate, :math:`x`. For numbers, arrays,
        and tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise hyperbolic tangent, :math:`\tanh{x}`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "tanh"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Log(Node):
    r"""
    Calculate the element-wise natural logarithm of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose natural logarithm you want to calculate, :math:`x`. For numbers, arrays,
        and tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise natural logarithm, :math:`\log{x}`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "log"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Exp(Node):
    r"""
    Calculate the element-wise exponential of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose exponential you want to calculate, :math:`x`. For numbers, arrays,
        and tensors, the object is converted to a tensor and then the operation is applied. For
        functions of time (Pwcs and Stfs), the composition of the operation with the function
        is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise exponential, :math:`e^{x}`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "exp"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Negative(Node):
    r"""
    Calculate the element-wise numerical negative value of an object. This can be a number,
    an array, a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose numerical negative value you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise negation, :math:`-x`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "negative"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Real(Node):
    r"""
    Calculate the element-wise real part of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose real part you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise real part, :math:`\Re(x)`, of the values or function you provided. The
        returned object is a real object of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.

    See Also
    --------
    Graph.complex_value : Create a complex object from its real and imaginary parts.
    Graph.imag : Imaginary part of a complex object.
    """

    name = "real"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Imag(Node):
    r"""
    Calculate the element-wise imaginary part of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose imaginary part you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise imaginary part, :math:`\Im(x)`, of the values or function you provided. The
        returned object is a real object of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.

    See Also
    --------
    Graph.complex_value : Create a complex object from its real and imaginary parts.
    Graph.real : Real part of a complex object.
    """

    name = "imag"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Absolute(Node):
    r"""
    Calculate the element-wise absolute value of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose absolute value you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise absolute value, :math:`\left|x\right|`, of the values or function you
        provided. The returned object is a real object of the same kind as the one you provided,
        except if you provide a number or an np.ndarray in which case it's a Tensor.

    See Also
    --------
    Graph.angle : Argument of a complex object.
    Graph.complex_value : Create a complex object from its real and imaginary parts.
    """

    name = "abs"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Angle(Node):
    r"""
    Calculate the element-wise argument of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose argument you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise argument, :math:`\arg(x)`, of the values or function you provided. The
        returned object is a real object of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.

    See Also
    --------
    Graph.abs : Absolute value of a complex object.
    Graph.complex_value : Create a complex object from its real and imaginary parts.
    """

    name = "angle"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Conjugate(Node):
    r"""
    Calculate the element-wise complex conjugate of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The object whose complex conjugate you want to calculate, :math:`x`. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is
        applied. For functions of time (Pwcs and Stfs), the composition of the operation with
        the function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise complex conjugate, :math:`x^\ast`, of the values or function you
        provided. The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.

    See Also
    --------
    Graph.adjoint : Hermitian adjoint of an operator.
    """

    name = "conjugate"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Arcsin(Node):
    r"""
    Calculate the element-wise arcsine of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : float or np.ndarray or Tensor or Pwc or Stf
        The object whose arcsine you want to calculate, :math:`x`. Must be real. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is applied.
        For functions of time (Pwcs and Stfs), the composition of the operation with the
        function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise arcsine, :math:`\arcsin{x}`, of the values or function you
        provided. Outputs will be in the range of :math:`[-\pi/2, \pi/2]`.
        The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "arcsin"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Arccos(Node):
    r"""
    Calculate the element-wise arccosine of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : float or np.ndarray or Tensor or Pwc or Stf
        The object whose arccosine you want to calculate, :math:`x`. Must be real. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is applied.
        For functions of time (Pwcs and Stfs), the composition of the operation with the
        function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise arccosine, :math:`\arccos{x}`, of the values or function you
        provided. Outputs will be in the range of :math:`[0, \pi]`.
        The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "arccos"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Arctan(Node):
    r"""
    Calculate the element-wise arctangent of an object. This can be a number, an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf.

    Parameters
    ----------
    x : float or np.ndarray or Tensor or Pwc or Stf
        The object whose arctangent you want to calculate, :math:`x`. Must be real. For numbers,
        arrays, and tensors, the object is converted to a tensor and then the operation is applied.
        For functions of time (Pwcs and Stfs), the composition of the operation with the
        function is computed (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise arctangent, :math:`\arctan{x}`, of the values or function you
        provided. Outputs will be in the range of :math:`[-\pi/2, \pi/2]`.
        The returned object is of the same kind as the one you provided, except if you
        provide a number or an np.ndarray in which case it's a Tensor.
    """

    name = "arctan"
    args = [forge.arg("x", type=NumericOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_unary_node_data(_operation, **kwargs)


class Adjoint(Node):
    r"""
    Calculate the element-wise adjoint of the last two dimensions of an object.
    This can be a an array, a tensor, or a time-dependent function in the form of a
    Pwc or an Stf where values have at least two dimensions.

    Parameters
    ----------
    x : np.ndarray or Tensor or Pwc or Stf
        The object whose adjoint you want to calculate, :math:`X^\dagger`.
        Must be a matrix or a matrix-valued function.
        For arrays and tensors, the object is converted to a tensor and then
        the operation is applied. For functions of time (Pwcs and Stfs), the composition
        of the operation with the function is computed
        (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise adjoint, of the last two dimension of the given matrix or matrix-valued
        function.

    See Also
    --------
    Graph.conjugate : Conjugate of a complex object.
    Graph.transpose : Reorder the dimensions of a tensor.
    """

    name = "adjoint"
    args = [forge.arg("x", type=TensorLikeOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        def value_shape_changer(value_shape: Tuple[int, ...]) -> Tuple[int, ...]:
            check_argument(
                len(value_shape) >= 2,
                "x must be at least 2D.",
                {"x": kwargs.get("x")},
                extras={"x.value_shape": value_shape},
            )
            # Unpacking syntax is awesome.
            *batch, x, y = value_shape
            return (*batch, y, x)

        return _create_flexible_unary_node_data(
            _operation,
            **kwargs,
            value_shape_changer=value_shape_changer,
            can_be_number=False,
        )


class Trace(Node):
    r"""
    Calculate the trace of an object. This can be a an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf
    where values have at least two dimensions.
    The trace is calculated on the last two dimensions.

    Parameters
    ----------
    x : np.ndarray or Tensor or Pwc or Stf
        The object whose trace you want to calculate, :math:`\mathop{\mathrm{Tr}}(x)`.
        Must be a matrix or a matrix-valued function.
        For arrays and tensors, the object is converted to a tensor and then
        the operation is applied. For functions of time (Pwcs and Stfs), the composition
        of the operation with the function is computed
        (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The trace of the last two dimension of the given matrix or matrix-valued
        function. Outputs will have two fewer dimensions.

    See Also
    --------
    Graph.einsum : Tensor contraction via Einstein summation convention.
    """

    name = "trace"
    args = [forge.arg("x", type=TensorLikeOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        def value_shape_changer(value_shape: Tuple[int, ...]) -> Tuple[int, ...]:
            check_argument(
                len(value_shape) >= 2,
                "x must be at least 2D.",
                {"x": kwargs.get("x")},
                extras={"x.value_shape": value_shape},
            )
            return value_shape[:-2]

        return _create_flexible_unary_node_data(
            _operation,
            **kwargs,
            value_shape_changer=value_shape_changer,
            can_be_number=False,
        )


class HermitianPart(Node):
    r"""
    Calculate the Hermitian part of an object. This can be an array,
    a tensor, or a time-dependent function in the form of a Pwc or an Stf
    where values have at least two dimensions.
    The operation is applied on the last two dimensions, which must be equal to each other.

    Parameters
    ----------
    x : np.ndarray or Tensor or Pwc or Stf
        The object whose Hermitian part you want to calculate, :math:`\mathop{x}`.
        Must be a matrix or a matrix-valued function.
        For arrays and tensors, the object is converted to a tensor and then
        the operation is applied. For functions of time (Pwcs and Stfs), the composition
        of the operation with the function is computed
        (that is, the operation is applied to the function values).
    name : str or None, optional
        The name of the node. You can only provide a name if the object is not an Stf.

    Returns
    -------
    Tensor or Pwc or Stf
        The Hermitian part of the matrix or matrix-valued function,
        :math:`\frac{1}{2}(\mathop{x}+\mathop{x}^\dagger)`.
        Outputs will have the same dimension as the inputs.

    See Also
    --------
    Graph.adjoint : Hermitian adjoint of an operator.

    Examples
    --------
    Create a Hamiltonian from a non-Hermitian Pwc operator.

    >>> omega = graph.pwc(durations=np.array([0.5, 0.7]), values=np.array([0.2, 0.4]))
    >>> sigma_m = np.array([[0, 1], [0, 0]])
    >>> operator = omega * sigma_m
    >>> graph.hermitian_part(operator, name="hamiltonian")
    <Pwc: name="hamiltonian", operation_name="hermitian_part", value_shape=(2, 2), batch_shape=()>
    >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
    >>> result["output"]["hamiltonian"]
    {
        'durations': array([0.5, 0.7]),
        'values': array([
            [[0. , 0.1], [0.1, 0. ]],
            [[0. , 0.2], [0.2, 0. ]]
            ]),
        'time_dimension': 0
    }

    See more examples in the `Simulate the dynamics of a single qubit using computational graphs
    <https://docs.q-ctrl.com/boulder-opal/tutorials/simulate-the-dynamics-of-a-single-qubit-
    using-computational-graphs>`_ tutorial.

    Create a Hamiltonian from a non-Hermitian Stf operator.

    >>> operator = stf_signal * np.array([[0, 1, 0], [0, 0, np.sqrt(2)], [0, 0, 0]])
    >>> hamiltonian = graph.hermitian_part(operator)
    >>> hamiltonian
    <Stf: operation_name="hermitian_part", value_shape=(3, 3), batch_shape=()>

    Create a Hermitian matrix from a non-Hermitian np.ndarray.

    >>> sigma_m = np.array([[0, 1], [0, 0]])
    >>> graph.hermitian_part(sigma_m, name="hamiltonian")
    <Tensor: name="hamiltonian", operation_name="hermitian_part", shape=(2, 2)>
    >>> result = bo.execute_graph(graph=graph, output_node_names="hamiltonian")
    >>> result["output"]["hamiltonian"]
    {'value': array([[0. , 0.5], [0.5, 0. ]])}
    """
    name = "hermitian_part"
    args = [forge.arg("x", type=TensorLikeOrFunction)]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        def value_shape_changer(value_shape: Tuple[int, ...]) -> Tuple[int, ...]:
            check_argument(
                len(value_shape) >= 2 and value_shape[-1] == value_shape[-2],
                "x must be a square matrix.",
                {"x": kwargs.get("x")},
                extras={"x.value.shape": value_shape},
            )
            return value_shape

        return _create_flexible_unary_node_data(
            _operation,
            **kwargs,
            value_shape_changer=value_shape_changer,
            can_be_number=False,
        )



from qctrlworkflowclient.router.api import DecodedResult


def metadata_formatter(input_: DecodedResult) -> DecodedResult:
    """
    Insert `action_id` to the `metadata` key of the result.

    Parameters
    ----------
    input_ : DecodedResult
        The result from the workflow functions.

    Returns
    -------
    DecodedResult
        The reformatted result.
    """
    result = input_.decoded
    assert isinstance(result, dict)
    if result.get("metadata") is None:
        result["metadata"] = {}
    result["metadata"].update({"action_id": input_.action_id})
    return input_


def async_result_formatter(input_: DecodedResult) -> dict:
    """
    Format the response result before updating it in the async result dictionary.

    Parameters
    ----------
    input_ : DecodedResult
        The result from the workflow functions.

    Returns
    -------
    dict
        The reformatted result.
    """
    result = metadata_formatter(input_)
    return result.decoded


def metadata_local_formatter(input_: dict) -> dict:
    """
    Fill action_id key as None in local mode.

    Parameters
    ----------
    input_ : dict
        Result from running workflow functions locally.

    Returns
    -------
    dict
        Formatted result.
    """
    if input_.get("metadata") is None:
        input_["metadata"] = {}
    input_["metadata"]["action_id"] = None
    return input_



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from dataclasses import asdict
from typing import Optional

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._core import reconstruct_noise_workflow
from boulderopal._validation.basic import (
    ArrayDType,
    nullable,
)
from boulderopal.noise_reconstruction._classes import (
    ALLOWED_NOISE_RECONSTRUCTION_METHODS,
    FilterFunction,
    NoiseReconstructionMethod,
    SVDEntropyTruncation,
    SVDFixedLengthTruncation,
)


def _check_filter_function_list(
    filter_functions_list: list[FilterFunction],
    filter_functions_name: str,
    control_count: int,
) -> None:
    """
    Check that all the filter functions for a given noise channel are valid.
    """
    check_argument(
        isinstance(filter_functions_list, list)
        and all(
            isinstance(filter_function, FilterFunction)
            for filter_function in filter_functions_list
        ),
        "The filter functions must be a list of lists of FilterFunction objects",
        {filter_functions_name: filter_functions_list},
    )
    check_argument(
        len(filter_functions_list) == control_count,
        "The list of filter functions for a noise channel must have the"
        " same length as the infidelities.",
        {filter_functions_name: filter_functions_list},
        extras={
            "len(infidelities)": control_count,
            f"len({filter_functions_name})": len(filter_functions_list),
        },
    )
    check_argument(
        all(
            filter_functions_list[0].sample_count == filter_function.sample_count
            for filter_function in filter_functions_list[1:]
        ),
        "All the filter functions associated with a noise channel must"
        " have the same number of samples.",
        {filter_functions_name: filter_functions_list},
    )
    check_argument(
        all(
            np.allclose(
                filter_functions_list[0].frequencies, filter_function.frequencies
            )
            for filter_function in filter_functions_list[1:]
        ),
        "The sample frequencies of all the filter functions associated with"
        " a noise channel must be the same.",
        {filter_functions_name: filter_functions_list},
    )


def _check_svd_fixed_length_truncation(
    method: SVDFixedLengthTruncation, control_count: int, total_sampled_frequencies: int
) -> None:
    """
    Perform additional validation for SVDFixedLengthTruncation.
    """
    if method.singular_value_count is not None:
        check_argument(
            method.singular_value_count <= control_count,
            "The singular value count for the SVD method with fixed length"
            " must be less than or equal to the number of infidelities.",
            {"method": method},
            extras={
                "singular_value_count": method.singular_value_count,
                "len(infidelities)": control_count,
            },
        )
        check_argument(
            method.singular_value_count <= total_sampled_frequencies,
            "The singular value count for the SVD method with fixed length"
            " must be less than or equal to the total number of frequencies sampled"
            " for the filter functions of each control.",
            {"method": method},
            extras={
                "singular_value_count": method.singular_value_count,
                "total sampled frequencies": total_sampled_frequencies,
            },
        )


def reconstruct(
    filter_functions: list[list[FilterFunction]],
    infidelities: np.ndarray,
    infidelity_uncertainties: Optional[np.ndarray] = None,
    method: Optional[NoiseReconstructionMethod] = None,
) -> dict:
    r"""
    Estimate the power spectral density (PSD) of noise processes affecting a quantum system.

    Use this function to obtain noise PSDs from measurements performed on
    your quantum system. You must provide the measurements as filter functions,
    which describe the controls applied to the system, and operational
    infidelities.

    Parameters
    ----------
    filter_functions : list[list[FilterFunction]]
        The filter functions associated with each control and noise. Each filter
        function represents the sensitivity of a set of controls to a certain
        noise. The placement in the outer list corresponds to the noise,
        while the inner list is organized by control. Note that at least one filter
        function must be present, and that the filter functions for each noise
        channel have to be sampled at the same frequencies.
    infidelities : np.ndarray
        The infidelities associated with the application of each control.
        It must contain at least one element, and all its values must be
        greater than or equal to 0, and less than or equal to 1.
    infidelity_uncertainties : np.ndarray or None, optional
        The uncertainty associated with each infidelity. The array must have the same
        length as the `infidelities`, and all values must be greater than or equal to
        0, and less than or equal to 1. Defaults to None, in which case no
        uncertainty is associated to the infidelities.
    method : NoiseReconstructionMethod or None, optional
        The method to be used in the noise reconstruction. Defaults to the
        singular value decomposition (SVD) method with entropy truncation at 0.5.

    Returns
    -------
    dict
        A dictionary containing the noise reconstruction result, with the following keys:

        ``output``
            A list with the spectral distributions of the reconstructed noises.
            Each list entry is a dictionary containing the power spectral densities of a noise
            (and the frequencies at which they are defined),
            presented in the same sequence as provided in the `filter_functions` argument.
            It might contain the estimated uncertainties for the spectral densities,
            if you provide infidelity uncertainties.

        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    Notes
    -----
    From the filter function theory [1]_, the operational infidelity for a given control
    sequence applied on a weak-noise-perturbed quantum system is the overlap between the
    noise power spectral density (PSD) and the corresponding filter functions:

    .. math::
        {\mathcal I}_j = \sum_{k = 1}^{N_{\mathrm{noise}}} \int {\mathrm d}f \,
        F_{jk}(f) S_k(f) ,

    where :math:`S_k(f)` is the PSD for the noise channel :math`k`, :math:`F_{jk}(f)` is
    the filter function corresponding to the control sequence :math:`j` and the noise
    channel :math:`k`, and :math:`{\mathcal I}_j` is the measured infidelity after the
    control :math:`j` is applied to the system. Discretizing the integrals for all
    :math:`M` measurements gives the following linear equation:

    .. math::
        F'{\mathbf S} = {\mathbf I} ,

    where :math:`F' = [F'_1, \dots, F'_j, \dots, F'_{N_{\mathrm{noise}}}]`
    is a :math:`M \times K` matrix and each element :math:`F'_j` is a
    :math:`M \times K_j` matrix representing the sampled filter functions weighted by
    discretized frequency step for the noise channel :math:`j` and
    :math:`K \equiv \sum_{j=1}^{N_\mathrm{noise}} K_j`; noise PSD
    :math:`{\mathbf S} = [{\mathbf S}_1, \dots, {\mathbf S}_j \dots, {\mathbf S}_K]^\top`
    is a :math:`K \times 1` vector and each element :math:`{\mathbf S}_j` is a
    :math:`K_j \times 1` vector for noise channel :math:`j`; infidelity vector
    :math:`{\mathbf I} = [{\mathcal I}_1, {\mathcal I}_2, \dots, {\mathcal I}_M]^\top`
    is a :math:`M \times 1` vector. Given sampled filter functions and infidelities,
    this function gives an estimation of the noise PSD :math:`{\mathbf S}_{\mathrm{est}}`.
    If uncertainties are provided with infidelities, this function also returns the
    uncertainties in estimation.

    References
    ----------
    .. [1] `T. J. Green, J. Sastrawan, H. Uys, and M. J. Biercuk,
            New Journal of Physics 15, 095004 (2013).
            <https://doi.org/10.1088/1367-2630/15/9/095004>`_

    Examples
    --------
    Perform a simple noise reconstruction of one noise affecting frequency 2,
    using two pulses perfectly sensitive to frequencies 1 and 2:

    >>> filter_functions = [
    ...    [
    ...        bo.noise_reconstruction.FilterFunction(
    ...            frequencies=np.array([1, 2]), inverse_powers=np.array([1, 0])
    ...        ),
    ...        bo.noise_reconstruction.FilterFunction(
    ...            frequencies=np.array([1, 2]), inverse_powers=np.array([0, 1])
    ...        )
    ...    ]
    ... ]
    >>> result = bo.noise_reconstruction.reconstruct(
    ...    filter_functions=filter_functions, infidelities=np.array([0, 1])
    ... )
    >>> result["output"]
        [{'frequencies': array([1., 2.]), 'psd': array([0., 1.])}]

    See also the `How to perform noise spectroscopy on arbitrary noise channels
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-
    perform-noise-spectroscopy-on-arbitrary-noise-channels>`_ user guide.
    """
    infidelities = ArrayDType.REAL(
        infidelities,
        "infidelities",
        ndim=1,
        min_=0,
        min_inclusive=True,
        max_=1,
        max_inclusive=True,
    )
    control_count = len(infidelities)
    check_argument(
        len(infidelities) > 1,
        "You must provide at least one infidelity.",
        {"infidelities": infidelities},
    )

    infidelity_uncertainties = nullable(
        ArrayDType.REAL,
        infidelity_uncertainties,
        "infidelity_uncertainties",
        ndim=1,
        shape=(control_count,),
        min_=0,
        min_inclusive=True,
        max_=1,
        max_inclusive=True,
    )

    check_argument(
        isinstance(filter_functions, list),
        "The filter functions must be passed as a list.",
        {"filter_functions": filter_functions},
        extras={"type(filter_functions)": type(filter_functions)},
    )
    check_argument(
        len(filter_functions) > 0,
        "The list of filter functions must have at least one element.",
        {"filter_functions": filter_functions},
    )

    # Generate two lists with noise_count elements each. One contains frequencies
    # as (noise_sample_count,) arrays; the other contains filter functions values,
    # as (sample_count, noise_sample_count) arrays.
    # Note that noise_sample_count can be different for different noises.
    noises_frequencies = []
    sampled_filter_functions = []
    for index, filter_functions_list in enumerate(filter_functions):
        _check_filter_function_list(
            filter_functions_list, f"filter_functions[{index}]", control_count
        )
        noises_frequencies.append(filter_functions_list[0].frequencies)
        sampled_filter_functions.append(
            np.array(
                [
                    filter_function.inverse_powers
                    for filter_function in filter_functions_list
                ]
            )
        )

    if method is None:
        method = SVDEntropyTruncation(rounding_threshold=0.5)

    check_argument(
        isinstance(method, ALLOWED_NOISE_RECONSTRUCTION_METHODS),
        "Unrecognized method type.",
        {"method": method},
        extras={"type(method)": type(method)},
    )

    # Additional validation for the Fixed-Length Truncation SVD.
    if isinstance(method, SVDFixedLengthTruncation):
        total_sampled_frequencies = sum(
            filter_function.sample_count for filter_function in filter_functions[:][0]
        )
        _check_svd_fixed_length_truncation(
            method=method,
            control_count=control_count,
            total_sampled_frequencies=total_sampled_frequencies,
        )

    return reconstruct_noise_workflow(
        method=asdict(method),
        noises_frequencies=noises_frequencies,
        filter_functions=sampled_filter_functions,
        infidelities=infidelities,
        infidelity_uncertainties=infidelity_uncertainties,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from boulderopal.ions._drives import (
    ComplexOptimizableDrive,
    Drive,
    OptimizableDrive,
    RealOptimizableDrive,
)
from boulderopal.ions._ion_chain_properties import obtain_ion_chain_properties
from boulderopal.ions._molmer_sorensen import (
    ms_optimize,
    ms_simulate,
)

__all__ = [
    "ComplexOptimizableDrive",
    "Drive",
    "OptimizableDrive",
    "RealOptimizableDrive",
    "ms_optimize",
    "ms_simulate",
    "obtain_ion_chain_properties",
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from dataclasses import dataclass
from enum import (
    Enum,
    auto,
)
from typing import (
    Dict,
    List,
)


@dataclass
class DocumentationSection:
    """
    A section of the graphs documentation.

    Parameters
    ----------
    title : str
        The title of the section.
    description : str
        The description of the section.
    operations : str
        The names of the operations to be listed in the section.
    subsections : list[DocumentationSection]
        The subsections that should be nested inside the section.
    """

    title: str
    description: str
    operations: List[str]
    subsections: List["DocumentationSection"]


class Category(Enum):
    """
    The different categories of graph operations.
    """

    ARITHMETIC_FUNCTIONS = auto()
    BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS = auto()
    BUILDING_SMOOTH_HAMILTONIANS = auto()
    DEPRECATED_OPERATIONS = auto()
    FILTERING_AND_DISCRETIZING = auto()
    LARGE_SYSTEMS = auto()
    LINEAR_ALGEBRA = auto()
    MATH = auto()
    MANIPULATING_TENSORS = auto()
    MOLMER_SORENSEN = auto()
    OPTIMAL_AND_ROBUST_CONTROL = auto()
    OPTIMIZATION_VARIABLES = auto()
    OTHER_OPERATIONS = auto()
    QUANTUM_INFORMATION = auto()
    RANDOM_OPERATIONS = auto()
    TIME_EVOLUTION = auto()
    SIGNALS = auto()


def create_documentation_sections(
    operations: Dict[str, List[Category]]
) -> List[DocumentationSection]:
    """
    Create sections containing the full operations documentation.

    Parameters
    ----------
    operations : dict[str, list[Category]]
        The operations and their associated categories.

    Returns
    -------
    list[DocumentationSection]
        A list of objects describing the top-level operations documentation sections. Each category
        appears exactly once in the nested structure of sections.
    """
    excluded_node_names = ["getattr", "getitem"]
    categories_map: Dict[Category, List[str]] = {category: [] for category in Category}

    for name, categories in sorted(operations.items()):
        for category in categories:
            if name not in excluded_node_names:
                categories_map[category].append(name)

    sections = [
        DocumentationSection(
            "Optimization variables",
            "When performing optimizations, you can use these operations to create the optimizable "
            "variables that can be tuned by the optimizer in order to minimize your cost function.",
            categories_map.pop(Category.OPTIMIZATION_VARIABLES),
            [],
        ),
        DocumentationSection(
            "Piecewise-constant tensor functions (PWCs)",
            "You can use these operations to create piecewise-constant functions "
            "either to represent control signals or system Hamiltonians.",
            categories_map.pop(Category.BUILDING_PIECEWISE_CONSTANT_HAMILTONIANS),
            [],
        ),
        DocumentationSection(
            "Sampleable tensor functions (STFs)",
            "You can use these functions to represent time-dependent functions in "
            "Boulder Opal.",
            categories_map.pop(Category.BUILDING_SMOOTH_HAMILTONIANS),
            [],
        ),
        DocumentationSection(
            "Filtering and discretizing",
            "You can use these functions to filter or resample the control signals.",
            categories_map.pop(Category.FILTERING_AND_DISCRETIZING),
            [],
        ),
        DocumentationSection(
            "Predefined signals",
            "You can use these operations to create common analytical signals. "
            "\n\nFor creating analytical signals not bound to a graph see the "
            ":ref:`Signal library` module.",
            categories_map.pop(Category.SIGNALS),
            [],
        ),
        DocumentationSection(
            "Quantum information",
            "You can use these operations to calculate common operations and metrics "
            "from quantum information theory.",
            categories_map.pop(Category.QUANTUM_INFORMATION),
            [],
        ),
        DocumentationSection(
            "Time evolution",
            "You can use these operations to calculate the time evolution of your open or closed "
            "quantum system, either for simulations or optimizations.",
            categories_map.pop(Category.TIME_EVOLUTION),
            [],
        ),
        DocumentationSection(
            "Optimal and robust control",
            "You can use these operations, together with the operations for creating "
            ":ref:`optimization variables<Optimization variables>` "
            "to define convenient cost functions for optimal and robust control.",
            categories_map.pop(Category.OPTIMAL_AND_ROBUST_CONTROL),
            [],
        ),
        DocumentationSection(
            "Large systems",
            "You can use these operations to build graphs that efficiently handle "
            "large quantum systems.",
            categories_map.pop(Category.LARGE_SYSTEMS),
            [],
        ),
        DocumentationSection(
            "Mølmer–Sørensen gates",
            "You can use these operations to efficiently model systems described by "
            "Mølmer–Sørensen interactions.",
            categories_map.pop(Category.MOLMER_SORENSEN),
            [],
        ),
        DocumentationSection(
            "Random operations",
            "You can use these operations to create random quantities, which take different values "
            "each time they are evaluated. These operations are most useful in simulations and "
            "stochastic optimizations.",
            categories_map.pop(Category.RANDOM_OPERATIONS),
            [],
        ),
        DocumentationSection(
            "Manipulating tensors",
            "You can use these operations to manipulate the structures of tensors.",
            categories_map.pop(Category.MANIPULATING_TENSORS),
            [],
        ),
        DocumentationSection(
            "Arithmetic", "", categories_map.pop(Category.ARITHMETIC_FUNCTIONS), []
        ),
        DocumentationSection(
            "Linear algebra", "", categories_map.pop(Category.LINEAR_ALGEBRA), []
        ),
        DocumentationSection(
            "Basic mathematical functions", "", categories_map.pop(Category.MATH), []
        ),
        DocumentationSection(
            "Other operations",
            "You typically do not need to use these operations directly.",
            categories_map.pop(Category.OTHER_OPERATIONS),
            [],
        ),
        DocumentationSection(
            "Deprecated operations",
            "These operations are deprecated and will be removed in the future.",
            categories_map.pop(Category.DEPRECATED_OPERATIONS),
            [],
        ),
    ]
    # Verify that we dealt with all the categories.
    assert not categories_map
    return sections



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from enum import Enum

from qctrlcommons.preconditions import check_argument

from boulderopal._core import execute_graph_workflow
from boulderopal._nodes.registry import PRIMARY_NODE_REGISTRY
from boulderopal._validation.basic import validate_enum
from boulderopal._validation.graph import validate_output_node_names
from boulderopal.graph._graph import Graph


class ExecutionMode(Enum):
    """
    Configuration for the execution mode used in `execute_graph`.

    Attributes
    ----------
    COMPILED
        Compiled execution, where the graph you provide is compiled into an executable that can
        run operations in parallel and with low overhead (at the expense of compilation time).
        This mode can be faster for simulations of open system dynamics, large systems, or systems
        described by sampleable tensor function (Stf) objects.

    EAGER
        Eager execution, where the graph nodes you provide are executed sequentially
        immediately as they are encountered. This mode can be faster for simulations of systems
        described by piecewise-constant Hamiltonians with a large number (more than roughly 1000)
        of segments.

    See Also
    --------
    boulderopal.execute_graph : Evaluate a graph corresponding to a set of generic functions.
    """

    COMPILED = "COMPILED"
    EAGER = "EAGER"


def execute_graph(
    graph: Graph,
    output_node_names: str | list[str],
    execution_mode: ExecutionMode = ExecutionMode.COMPILED,
) -> dict:
    r"""
    Evaluate a graph corresponding to a set of generic functions.

    Use this function to carry out computations expressed as a graph
    representing a collection of arbitrary functions.

    Parameters
    ----------
    graph : Graph
        The graph describing the outputs. It must contain nodes with names
        (giving the output functions).
    output_node_names : str or list[str]
        The names of the graph nodes that define the output functions.
        The function evaluates these and returns them in the output.
        You can pass a string for a single node or a list of node names.
    execution_mode : ExecutionMode, optional
        The execution mode to use for the calculation. Choosing a custom execution mode can lead to
        faster computations in certain cases. Defaults to compiled execution mode.

    Returns
    -------
    dict
        A dictionary containing the graph execution result, with the following keys:

        ``output``
            The dictionary giving the value of each requested output node.
            The keys of the dictionary are the names of the output nodes.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    boulderopal.closed_loop.optimize :
        Run a closed-loop optimization to find a minimum of the given cost function.
    boulderopal.closed_loop.step : Perform a single step in a closed-loop optimization.
    boulderopal.run_gradient_free_optimization :
        Perform model-based optimization without using gradient values.
    boulderopal.run_optimization :
        Perform gradient-based deterministic optimization of generic real-valued functions.
    boulderopal.run_stochastic_optimization :
        Perform gradient-based stochastic optimization of generic real-valued functions.

    Notes
    -----
    This function computes arbitrary functions represented by a graph.

    The graph is made up of primitive nodes, where each node represents
    a function of the output of the nodes (or constant values) it depends on.
    You can assign a name to any node and request a list of named nodes :math:`\{s_j\}`
    as the outputs to be evaluated.

    Examples
    --------
    See the `How to represent quantum systems using graphs
    <https://docs.q-ctrl.com/boulder-opal/user-guides/
    how-to-represent-quantum-systems-using-graphs>`_ user guide.
    """

    for operation in graph.operations.values():
        check_argument(
            not PRIMARY_NODE_REGISTRY.get_node_cls(
                operation.operation_name
            ).optimizable_variable,
            f"The operation '{operation.operation_name}' is not available to use"
            " as part of the execute_graph function.",
            {"graph": graph},
        )

    output_node_names = validate_output_node_names(output_node_names, graph)
    return execute_graph_workflow(
        graph, output_node_names, validate_enum(ExecutionMode, execution_mode)
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass

from qctrlcommons.node.base import Node


@dataclass
class _PrimaryNamespaceDoc:
    """
    A helper class for building the doc for the primary node namespaces.
    """

    namespace_name: str
    sub_nodes: dict
    categories: list


def create_namespace_doc(
    namespace_name, module_name, categories, node_name_formatter=lambda x: x
):
    """
    Create a _PrimaryNamespaceDoc given a namespace name, the name of the module
    and a list of categories.
    """
    # Collect the `sub_nodes`.
    sub_nodes = {}
    for _, obj in inspect.getmembers(sys.modules[module_name]):
        if (
            inspect.isclass(obj)
            and obj.__module__ == module_name
            and issubclass(obj, Node)
        ):
            new_name = node_name_formatter(obj.name)
            sub_nodes[new_name] = obj.create_graph_method()
    return _PrimaryNamespaceDoc(namespace_name, sub_nodes, categories)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
    nullable,
)
from boulderopal.superconducting._drives import (
    OptimizableCoefficient,
    RealOptimizableConstant,
    RealOptimizableSignal,
)

RealAtype = TypeVar(
    "RealAtype", int, float, np.ndarray, RealOptimizableConstant, RealOptimizableSignal
)
RealBtype = TypeVar(
    "RealBtype", int, float, np.ndarray, RealOptimizableConstant, RealOptimizableSignal
)
Coefficient = TypeVar(
    "Coefficient", int, float, complex, np.ndarray, OptimizableCoefficient
)


def validate_real_coefficient(value: Any, name: str) -> Any:
    """
    Ensure coefficient is real-valued.
    """
    if isinstance(value, (RealOptimizableSignal, RealOptimizableConstant)):
        return value
    try:
        try:
            return ScalarDType.REAL(value, name)
        except QctrlArgumentsValueError:
            return ArrayDType.REAL(value, name)
    except QctrlArgumentsValueError as err:
        # re-raise to throw a better error message.
        raise QctrlArgumentsValueError(
            f"The {name} must be real-valued. The supported types are scalar, array, "
            "RealOptimizableConstant, and RealOptimizableSignal",
            {name: value},
        ) from err


def _validate_coefficient(value: Any, name: str) -> Any:
    if isinstance(value, OptimizableCoefficient):
        return value
    try:
        try:
            return ScalarDType.NUMBER(value, name)
        except QctrlArgumentsValueError:
            try:
                return ArrayDType.REAL(value, name)
            except QctrlArgumentsValueError:
                return ArrayDType.COMPLEX(value, name)
    except QctrlArgumentsValueError as err:
        # re-raise to throw a better error message.
        raise QctrlArgumentsValueError(
            f"The {name} must be a number or OptimizableCoefficient.", {name: value}
        ) from err


@dataclass
class Transmon(Generic[RealAtype, RealBtype, Coefficient]):
    """
    Class that stores all the physical system data for a transmon.

    Parameters
    ----------
    dimension : int
        Number of dimensions of the Hilbert space of the transmon.
        Must be at least 2.
    frequency : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant or None, \
                optional
        The frequency of the transmon, :math:`\\omega_t`.
        If not provided, it defaults to no frequency term.
    anharmonicity : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant \
                or None, optional
        The nonlinearity of the transmon, :math:`\\alpha`.
        If not provided, it defaults to no anharmonicity term.
    drive : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant or None, optional
        The complex drive of the transmon, :math:`\\gamma_t`.
        If not provided, it defaults to no drive term.
    name : str, optional
        The identifier of the transmon that is used to link interaction terms to this transmon.
        Defaults to "transmon".

    See Also
    --------
    boulderopal.superconducting.Cavity :
        Class describing cavities in superconducting systems.
    boulderopal.superconducting.TransmonCavityInteraction :
        Class describing interactions between a transmon and a cavity.
    boulderopal.superconducting.TransmonTransmonInteraction :
        Class describing interactions between two transmons.

    Notes
    -----
    The Hamiltonian for the transmon is defined as

    .. math::
        H_\\mathrm{transmon} =
            \\omega_t b^\\dagger b
            + \\frac{\\alpha}{2} (b^\\dagger)^2 b^2
            + \\frac{1}{2} \\left(\\gamma_t b^\\dagger + H.c. \\right) ,

    where :math:`H.c.` indicates the Hermitian conjugate.
    All coefficients in the Hamiltonian are optional,
    and you should only pass those relevant to your system.
    """

    dimension: int
    frequency: Optional[RealAtype] = None
    anharmonicity: Optional[RealBtype] = None
    drive: Optional[Coefficient] = None
    name: str = "transmon"

    def __post_init__(self) -> None:
        self.dimension = ScalarDType.INT(
            self.dimension, "dimension", min_=2, min_inclusive=True
        )
        self.frequency = nullable(
            validate_real_coefficient, self.frequency, "frequency"
        )
        self.anharmonicity = nullable(
            validate_real_coefficient, self.anharmonicity, "anharmonicity"
        )
        self.drive = nullable(_validate_coefficient, self.drive, "drive")


@dataclass
class Cavity(Generic[RealAtype, RealBtype, Coefficient]):
    """
    Class that stores all the physical system data for a cavity.

    Parameters
    ----------
    dimension : int
        Number of dimensions of the Hilbert space of the cavity.
        Must be at least 2.
    frequency : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant, optional
        The frequency of the cavity or None, :math:`\\omega_c`.
        If not provided, it defaults to no frequency term.
    kerr_coefficient : real or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or None, optional
        The nonlinearity of the cavity, :math:`K`.
        If not provided, it defaults to no nonlinear term.
    drive : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant or None, optional
        The complex drive of the cavity, :math:`\\gamma_c`.
        If not provided, it defaults to no drive term.
    name : str, optional
        The identifier of the cavity that is used to link interaction terms to this cavity.
        Defaults to "cavity".

    See Also
    --------
    boulderopal.superconducting.CavityCavityInteraction :
        Class describing interactions between two cavities.
    boulderopal.superconducting.Transmon :
        Class describing transmons in superconducting systems.
    boulderopal.superconducting.TransmonCavityInteraction :
        Class describing interactions between a transmon and a cavity.

    Notes
    -----
    The Hamiltonian for the cavity is defined as

    .. math::
        H_\\mathrm{cavity} =
            \\omega_c a^\\dagger a
            + \\frac{K}{2} (a^\\dagger)^2 a^2
            + \\frac{1}{2} \\left(\\gamma_c a^\\dagger + H.c. \\right) ,

    where :math:`H.c.` indicates the Hermitian conjugate.
    All coefficients in the Hamiltonian are optional,
    and you should only pass those relevant to your system.
    """

    dimension: int
    frequency: Optional[RealAtype] = None
    kerr_coefficient: Optional[RealBtype] = None
    drive: Optional[Coefficient] = None
    name: str = "cavity"

    def __post_init__(self) -> None:
        self.dimension = ScalarDType.INT(
            self.dimension, "dimension", min_=2, min_inclusive=True
        )
        self.frequency = nullable(
            validate_real_coefficient, self.frequency, "frequency"
        )
        self.kerr_coefficient = nullable(
            validate_real_coefficient, self.kerr_coefficient, "kerr_coefficient"
        )
        self.drive = nullable(_validate_coefficient, self.drive, "drive")


@dataclass
class TransmonTransmonInteraction(Generic[Coefficient]):
    """
    Class that stores all the physical system data for the interaction
    between two transmons.

    Parameters
    ----------
    transmon_names : tuple[str, str]
        The two names identifying the transmons in the interaction.
    effective_coupling : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant or None, optional
        The effective coupling between the two transmons, :math:`g`.
        If not provided, it defaults to no effective coupling term.

    See Also
    --------
    boulderopal.superconducting.Transmon :
        Class describing transmons in superconducting systems.

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math::
        H_\\mathrm{transmon-transmon} = g b_1 b_2^\\dagger + H.c. .
    """

    transmon_names: tuple[str, str]
    effective_coupling: Optional[Coefficient]

    def __post_init__(self) -> None:
        check_argument(
            self.transmon_names[0] != self.transmon_names[1],
            "The names of the two transmons must be different.",
            {"transmon_names": self.transmon_names},
        )
        self.effective_coupling = nullable(
            _validate_coefficient, self.effective_coupling, "effective_coupling"
        )


@dataclass
class TransmonCavityInteraction(Generic[RealAtype, Coefficient]):
    """
    Class that stores all the physical system data for the interaction
    between a transmon and a cavity.

    Parameters
    ----------
    dispersive_shift : real or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or None, optional
        The dispersive shift between the transmon and the cavity, :math:`\\chi`.
        You must provide either a dispersive shift or a Rabi coupling.
    rabi_coupling : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant or None, optional
        The strength of the Rabi coupling between the transmon and the cavity, :math:`\\Omega`.
        You must provide either a dispersive shift or a Rabi coupling.
    transmon_name : str, optional
        The name identifying the transmon in the interaction.
        Defaults to "transmon".
    cavity_name : str, optional
        The name identifying the cavity in the interaction.
        Defaults to "cavity".

    See Also
    --------
    boulderopal.superconducting.Cavity :
        Class describing cavities in superconducting systems.
    boulderopal.superconducting.Transmon :
        Class describing transmons in superconducting systems.

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math:: H_\\mathrm{transmon-cavity} = \\chi a^\\dagger a b^\\dagger b ,

    or as

    .. math:: H_\\mathrm{transmon-cavity} = \\Omega a b^\\dagger + H.c. ,

    where :math:`H.c.` indicates the Hermitian conjugate.
    """

    dispersive_shift: Optional[RealAtype] = None
    rabi_coupling: Optional[Coefficient] = None
    transmon_name: str = "transmon"
    cavity_name: str = "cavity"

    def __post_init__(self) -> None:
        check_argument(
            (self.dispersive_shift is None) ^ (self.rabi_coupling is None),  # xor
            "You must provide either a dispersive shift or a Rabi coupling.",
            {
                "dispersive_shift": self.dispersive_shift,
                "rabi_coupling": self.rabi_coupling,
            },
        )
        self.dispersive_shift = nullable(
            validate_real_coefficient, self.dispersive_shift, "dispersive_shift"
        )
        self.rabi_coupling = nullable(
            _validate_coefficient, self.rabi_coupling, "rabi_coupling"
        )


@dataclass
class CavityCavityInteraction(Generic[RealAtype]):
    r"""
    Class that stores all the physical system data for the interaction
    between two cavities.

    Parameters
    ----------
    cavity_names : tuple[str, str]
        The two names identifying the cavities in the interaction.
    cross_kerr_coefficient : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant
        The cross-Kerr coefficient between the two cavities, :math:`K_{12}`.
        If not provided, it defaults to no cross-Kerr term.

    See Also
    --------
    boulderopal.superconducting.Cavity :
        Class describing cavities in superconducting systems.

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math::
        H_\mathrm{cavity-cavity} = K_{12} a_1^\dagger a_1 a_2^\dagger a_2 .
    """

    cavity_names: tuple[str, str]
    cross_kerr_coefficient: RealAtype

    def __post_init__(self) -> None:
        check_argument(
            self.cavity_names[0] != self.cavity_names[1],
            "The names of the two cavities must be different.",
            {"cavity_names": self.cavity_names},
        )
        self.cross_kerr_coefficient = validate_real_coefficient(
            self.cross_kerr_coefficient, "cross_kerr_coefficient"
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Callable,
    List,
    Tuple,
)

import numpy as np
from qctrlcommons.node.wrapper import (
    NamedNodeData,
    NodeData,
)


class ArithmeticMixin:
    """
    Mixin to be used by NodeData that support binary arithmetic operations with
    number/array/Tensor/Pwc/Stf objects.

    By default ``arr + graph.op()`` throws an error, since NumPy doesn't know how to add
    `graph.op()` objects to arrays. Even the fact that the func ops override `__radd__` doesn't
    help, since the NumPy addition takes precedence. We can instead tell NumPy to delegate all
    binary operations to the other operand, by explicitly clearing the `__array_ufunc__`
    attribute.
    """

    __array_ufunc__ = None

    def __add__(self, other):
        return self.operation.graph.add(self, other)

    def __radd__(self, other):
        return self.operation.graph.add(other, self)

    def __sub__(self, other):
        return self.operation.graph.subtract(self, other)

    def __rsub__(self, other):
        return self.operation.graph.subtract(other, self)

    def __matmul__(self, other):
        return self.operation.graph.matmul(self, other)

    def __rmatmul__(self, other):
        return self.operation.graph.matmul(other, self)

    def __mul__(self, other):
        return self.operation.graph.multiply(self, other)

    def __rmul__(self, other):
        return self.operation.graph.multiply(other, self)

    def __floordiv__(self, other):
        return self.operation.graph.floordiv(self, other)

    def __rfloordiv__(self, other):
        return self.operation.graph.floordiv(other, self)

    def __pow__(self, power):
        return self.operation.graph.pow(self, power)

    def __rpow__(self, other):
        return self.operation.graph.pow(other, self)

    def __truediv__(self, other):
        return self.operation.graph.truediv(self, other)

    def __rtruediv__(self, other):
        return self.operation.graph.truediv(other, self)

    def __abs__(self):
        return self.operation.graph.abs(self)

    def __neg__(self):
        return self.operation.graph.negative(self)


@dataclass
class Tensor(NamedNodeData, ArithmeticMixin):
    """
    A multi-dimensional array of data.

    Most functions accepting a :obj:`.Tensor` object can alternatively accept a NumPy array.

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Tensor` objects.

    Attributes
    ----------
    shape : tuple
        The shape of the tensor.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.tensor : Create a real or complex Tensor with the data provided.

    Notes
    -----
    The value of a `Tensor` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a dictionary
    with the "value" of the Tensor as a NumPy array.
    """

    shape: Tuple[int, ...]

    def __post_init__(self):
        self.operation.is_scalar_tensor = np.prod(self.shape) == 1

    def __getitem__(self, item) -> "Operation":
        """
        refer to item in operation.

        Returns
        -------
        Operation
            getitem operation.
        """
        node_data = self.operation.graph.getitem(self, item)
        shape = np.empty(self.shape)[item].shape
        return Tensor(node_data.operation, shape=shape)

    def __iter__(self):
        # Disable iteration for now. Even though this should work fine in theory (since all client
        # tensors have fully-defined shapes), allowing iterability on the client causes tensors to
        # pass checks that will fail in the backend (for example, if tensors are iterable on the
        # client, a multi-dimensional tensor can be passed to a function that expects a list of
        # tensors; such an input will fail in the backend though). This could be revisited in the
        # future if we're more strict about client-side validation of iterable inputs, or if we
        # update the backend to be able to iterate over tensors.
        raise TypeError(
            "You cannot iterate over Tensors directly. Instead you can iterate over the indices "
            "and extract elements of the tensor using `tensor[index]`."
        )

    def __repr__(self):
        return (
            f'<Tensor: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"shape={self.shape}>"
        )


@dataclass
class Pwc(NamedNodeData, ArithmeticMixin):
    """
    A piecewise-constant tensor-valued function of time (or batch of such functions).

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Pwc` objects or between a `Pwc` and a `Tensor`.

    Attributes
    ----------
    values : Tensor
        The values of the function on the piecewise-constant segments.
    durations : np.ndarray
        The durations of the constant segments.
    value_shape : tuple
        The shape of the function value.
    batch_shape : tuple
        The shape of the batch in the function.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.pwc : Operation to create piecewise-constant functions.

    Notes
    -----
    The value of a `Pwc` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a list of dictionaries,
    with the "duration" and "value" of the piecewise-constant function at each segment.
    Batch dimensions of the function will be represented as outer lists.

    For more information on `Pwc` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    value_shape: Tuple[int, ...]
    durations: np.ndarray
    batch_shape: Tuple[int, ...]

    @property
    def values(self):
        """
        Access to the values in Pwc.
        """
        node_data = self.operation.graph.getattr(self, "values")
        shape = (
            tuple(self.batch_shape) + (len(self.durations),) + tuple(self.value_shape)
        )
        return Tensor(node_data.operation, shape=shape)

    def __repr__(self):
        return (
            f'<Pwc: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}, batch_shape={self.batch_shape}>"
        )


@dataclass
class Stf(NodeData, ArithmeticMixin):
    """
    A sampleable tensor-valued function of time (or batch of such functions).

    You can use the arithmetic operators ``+``, ``-``, ``*``, ``**``, ``/``, ``//``, and ``@``
    to perform operations between two `Stf` objects or between an `Stf` and a `Tensor`.

    Attributes
    ----------
    value_shape : tuple
        The shape of the function value.
    batch_shape : tuple
        The shape of the batch in the function.

    See Also
    --------
    Graph.identity_stf : Operation to create an `Stf` representing the identity function.

    Notes
    -----
    Stf nodes represent arbitrary functions of time. Piecewise-constant (PWC) or constant functions
    are special cases of Stfs and the Q-CTRL Python package provides specific APIs to support them.
    Note that as the PWC property can simplify the calculation, you should always consider using
    PWC-related APIs if your system parameters or controls are described by PWC functions.

    The value of `Stf` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.

    For more information on `Stf` nodes see the `Working with time-dependent functions in
    Boulder Opal <https://docs.q-ctrl.com/boulder-opal/topics/working-with-time-dependent-
    functions-in-boulder-opal>`_ topic.
    """

    value_shape: Tuple[int, ...]
    batch_shape: Tuple[int, ...]

    def __repr__(self):
        return (
            f'<Stf: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}, batch_shape={self.batch_shape}>"
        )


@dataclass
class SparsePwc(NodeData):
    """
    A piecewise-constant sparse-matrix-valued function of time.

    Attributes
    ----------
    value_shape : tuple
        The shape of the function value.
    durations : np.ndarray
        The durations of the constant segments.

    See Also
    --------
    Graph.sparse_pwc_operator : Operation to create `SparsePwc` operators.

    Notes
    -----
    The value of `SparsePwc` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    value_shape: Tuple[int, ...]
    durations: np.ndarray

    def __repr__(self):
        return (
            f'<SparsePwc: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )


@dataclass
class Target(NodeData):
    """
    A target gate for an infidelity calculation.

    Attributes
    ----------
    value_shape : tuple
        The shape of the target gate operation.

    See Also
    --------
    Graph.target : Operation to define the target operation of a time evolution.

    Notes
    -----
    The value of `Target` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    value_shape: Tuple[int, ...]

    def __repr__(self):
        return (
            f'<Target: operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )


@dataclass
class ConvolutionKernel(NodeData):
    """
    A kernel to be used in a convolution.

    See Also
    --------
    Graph.convolve_pwc : Operation to create an `Stf` by convolving a `Pwc` with a kernel.
    Graph.gaussian_convolution_kernel : Operation to create a convolution kernel representing a
        normalized Gaussian.
    Graph.sinc_convolution_kernel : Operation to create a convolution kernel representing the sinc
        function.

    Notes
    -----
    The value of `ConvolutionKernel` nodes is not fetchable from graphs.
    Therefore, they do not have a `name` attribute.
    """

    def __repr__(self):
        return f'<ConvolutionKernel: operation_name="{self.operation.operation_name}">'


@dataclass
class Sequence(NamedNodeData):
    """
    Wrapper class for creating a sequence of Nodes.
    """

    items: List = field(default_factory=list)

    def create_sequence(self, node_constructor: Callable, size: int) -> "Sequence":
        """
        Populate the `items` of the sequence from the operation.

        Parameters
        ----------
        node_constructor : Callable
            A callable to generate the node data for the sequence.
        size : int
            Size of the sequence.

        Returns
        -------
        Sequence
            The sequence itself.
        """

        def get_item_op(item):
            return self.operation.graph.getitem(self, item).operation

        self.items = [
            node_constructor(get_item_op(index), index) for index in range(size)
        ]

        return self

    def __getitem__(self, index) -> "Operation":
        return self.items[index]

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return (
            f'<Sequence: operation_name="{self.operation.operation_name}", '
            f"items={self.items}>"
        )


@dataclass
class FilterFunction(NamedNodeData):
    """
    A tensor-valued filter function result.

    Attributes
    ----------
    inverse_powers : Tensor
        The values of the filter function at the given frequencies.
    uncertainties : Tensor, optional
        The uncertainties of the filter function values.
        This field is None when the exact method is used for computing the filter function.
    frequencies : np.ndarray
        The frequencies at which the filter function has been evaluated.
    value_shape : tuple[int]
        The shape of the function value.
    exact : bool
        Indicates whether filter function is exact.
    name : str
        The name assigned to the node.

    See Also
    --------
    Graph.filter_function :
        Evaluate the filter function for a control Hamiltonian and a noise operator at the given
        frequency elements.

    Notes
    -----
    The value of a `FilterFunction` node can be fetched in a graph calculation
    by adding its `name` in the `output_node_names` parameter for the function call.
    This will add an item to the output dictionary in the calculation result object,
    whose key is `name`. The item's value will be a dictionary
    with NumPy arrays with the "frequencies" and "inverse_powers" values.
    If the filter function calculation is not exact, the dictionary also
    contains an array with the filter function "uncertainties".
    """

    frequencies: np.ndarray
    exact: bool

    @property
    def value_shape(self):
        """
        Access the value shape.
        """
        return (len(self.frequencies),)

    @property
    def inverse_powers(self):
        """
        Access to the inverse powers in FilterFunction.
        """
        node_data = self.operation.graph.getattr(self, "inverse_powers")
        return Tensor(node_data.operation, shape=self.value_shape)

    @property
    def uncertainties(self):
        """
        Access to the uncertainties in FilterFunction.
        """
        if self.exact is False:
            node_data = self.operation.graph.getattr(self, "uncertainties")
            return Tensor(node_data.operation, shape=self.value_shape)
        return None

    def __repr__(self):
        return (
            f'<FilterFunction: name="{self.name}", '
            f'operation_name="{self.operation.operation_name}", '
            f"value_shape={self.value_shape}>"
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Any

from qctrlworkflowclient import (
    ApiRouter,
    CoreClientSettings,
    LocalRouter,
    Product,
    get_authenticated_client_for_product,
    get_default_api_url,
    get_default_cli_auth,
)
from qctrlworkflowclient.globals import global_value

from boulderopal._configuration.verbosity import MESSAGE_PRINTER
from boulderopal._constants import INVALID_SUBSCRIPTION_ERROR


def get_default_router() -> ApiRouter:
    """
    Return the default router that the Boulder Opal client uses.
    """
    client = get_authenticated_client_for_product(
        product_access_required="boulder-opal-cli-access",
        api_url=get_default_api_url(),
        auth=get_default_cli_auth(),
        invalid_access_error_message=INVALID_SUBSCRIPTION_ERROR,
    )
    settings = get_configuration()

    return ApiRouter(client, settings)


@global_value("BOULDER_OPAL_CONFIGURATION")
def get_configuration() -> CoreClientSettings:
    """
    Return the global Boulder Opal settings.
    """
    return CoreClientSettings(
        router=get_default_router,
        product=Product.BOULDER_OPAL,
        event_listeners=[MESSAGE_PRINTER],
    )


def configure(**kwargs: Any) -> None:
    """
    Update the global Boulder Opal settings. See :class:`CoreClientSettings`
    for details on which fields can be updated.
    """
    configuration = get_configuration()
    configuration.update(**kwargs)


def set_local(resolver: "BaseResolver") -> None:  # type: ignore
    """
    Configure Boulder Opal for local routing.

    Parameters
    ----------
    resolver : BaseResolver
        A local implementation of a workflow resolver which uses
        a registry that implements all of the available Boulder Opal
        workflows.
    """
    configure(router=LocalRouter(resolver))


def in_local_mode() -> bool:
    """
    Check if client is in local mode.
    """
    return isinstance(get_configuration().get_router(), LocalRouter)


def is_async(result: Any) -> bool:
    """
    Check if we get async result back from the server.
    """
    return isinstance(result, dict) and result.get("async_result") is not None



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utilities for the signal library nodes.
"""

from __future__ import annotations

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._nodes.node_data import Tensor


def validate_optimizable_parameter(
    graph: "Graph", parameter: float | Tensor, name: str
) -> Tensor:
    """
    Convert parameter into a Tensor, checks that it contains a single element
    and returns a Tensor containing that element.

    Parameters
    ----------
    graph : Graph
        Boulder computational graph object.
    parameter : float or Tensor
        The parameter to be validated.
    name : str
        The name of the parameter.

    Returns
    -------
    Tensor
        A Tensor object.
    """

    parameter = graph.tensor(parameter)
    check_argument(
        np.prod(parameter.shape) == 1,
        f"If passed as a Tensor, the {name} must either "
        "be a scalar or contain a single element.",
        {name: parameter},
        extras={f"{name}.shape": parameter.shape},
    )

    return graph.reshape(parameter, [1])[0]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for NodeRegistry."""
import inspect
from functools import partial

from boulderopal._nodes.composite import (
    filter_and_resample,
    optimizable_scalar,
    optimizable_signals,
    signals,
    signals_pwc,
    signals_stf,
)
from boulderopal._nodes.composite.namespace import (
    COMPOSITE_ATTR,
    NAMESPACE_ATTR,
)


def _register(modules):
    """
    Collect exposed composite nodes from modules.
    """
    registered = []
    for module in modules:
        for _, member in inspect.getmembers(
            module, predicate=partial(_filter, module_name=module.__name__)
        ):
            registered.append(member)
    return registered


def _filter(member, module_name):
    """
    - Exposed member must be a function.
    - Exposed member must be defined in the module, not imported.
    - Exposed member must not be private.
    """
    return (
        (inspect.isfunction(member) or inspect.isclass(member))
        and member.__module__ == module_name
        and not member.__name__.startswith("_")
        and hasattr(member, COMPOSITE_ATTR)
    )


_composite_registry = _register(
    [
        optimizable_signals,
        filter_and_resample,
        optimizable_scalar,
        signals_pwc,
        signals_stf,
        signals,
    ]
)

# Collect the composite node namespaces.
COMPOSITE_NAMESPACE_REGISTRY = [
    node for node in _composite_registry if hasattr(node, NAMESPACE_ATTR)
]

# Collect the composite nodes that are not in a namespace.
COMPOSITE_NODE_REGISTRY = [
    node for node in _composite_registry if not hasattr(node, NAMESPACE_ATTR)
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from boulderopal.superconducting._components import (
    Cavity,
    CavityCavityInteraction,
    Transmon,
    TransmonCavityInteraction,
    TransmonTransmonInteraction,
)
from boulderopal.superconducting._drives import (
    ComplexOptimizableConstant,
    ComplexOptimizableSignal,
    OptimizableCoefficient,
    RealOptimizableConstant,
    RealOptimizableSignal,
)
from boulderopal.superconducting._functions import (
    optimize,
    simulate,
)

__all__ = [
    "Cavity",
    "CavityCavityInteraction",
    "ComplexOptimizableConstant",
    "ComplexOptimizableSignal",
    "OptimizableCoefficient",
    "RealOptimizableConstant",
    "RealOptimizableSignal",
    "Transmon",
    "TransmonCavityInteraction",
    "TransmonTransmonInteraction",
    "optimize",
    "simulate",
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from functools import partial
from typing import Callable

from qctrlworkflowclient import core_workflow
from qctrlworkflowclient.router.api import DecodedResult

from boulderopal._configuration.configuration import (
    get_configuration,
    in_local_mode,
    is_async,
)
from boulderopal._core.formatter import (
    metadata_formatter,
    metadata_local_formatter,
)


def _formatter(
    result: DecodedResult | dict,
    formatters: tuple[Callable[[DecodedResult], DecodedResult], ...] = (
        metadata_formatter,
    ),
) -> dict:
    if result is None or (isinstance(result, DecodedResult) and result.decoded is None):
        raise RuntimeError("All workflow function should return a non-nullable result.")
    if in_local_mode():
        return metadata_local_formatter(result)
    if is_async(result):
        return result
    for _func in formatters:
        result = _func(result)
    assert isinstance(result, DecodedResult)
    return result.decoded


boulder_opal_workflow = partial(core_workflow, get_configuration, formatter=_formatter)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Union

import forge
import numpy as np
from qctrlcommons.exceptions import QctrlException
from qctrlcommons.node.base import Node
from qctrlcommons.preconditions import (
    check_argument,
    check_numeric_numpy_array,
)

from .arithmetic_unary import check_operand_type
from .documentation import Category
from .node_data import (
    Pwc,
    Stf,
    Tensor,
)
from .types import (
    NumericOrFunction,
    TensorLikeOrFunction,
)
from .utils import (
    get_broadcasted_shape,
    mesh_pwc_durations,
    validate_batch_and_value_shapes,
    validate_broadcasted_shape,
    validate_function_output_shapes,
    validate_shape,
    validate_tensor_and_function_output_shapes,
)


def _create_flexible_binary_node_data(
    _operation,
    op_name,
    x,
    y,
    name,
    validate_value_shape=validate_broadcasted_shape,
    can_x_be_number=True,
    can_y_be_number=True,
):
    """
    Common implementation of `create_node_data` for nodes acting on Tensors, Pwcs, and Stfs
    implementing binary functions.

    Parameters
    ----------
    _operation : Operation
        The operation to implement.
    op_name : str
        The name of the operation.
    x : number or np.ndarray or Tensor or Pwc or Stf
        The left operand.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The right operand.
    name : str
        The name of the node.
    validate_value_shape : Callable[[tuple, tuple, str, str], tuple], optional
        Function that takes the value shapes of two Tensors, Pwcs,
        or Stfs (as well as their names), and returns the expected values
        shape of the output Tensor, Pwc, or Stf. The function
        shouldn't assume that the shapes are compatible, and raise an
        exception if they aren't. The names provided should be used to
        generate the error message.
    can_x_be_number : bool, optional
        Whether x can be a number or not. Defaults to True.
    can_y_be_number : bool, optional
        Whether y can be a number or not. Defaults to True.

    Returns
    -------
    Tensor or Pwc or Stf
        The operation acting on the object.
    """
    check_operand_type(x, "x", can_be_number=can_x_be_number)
    check_operand_type(y, "y", can_be_number=can_y_be_number)
    # operation(Pwc, Stf) or operation(Stf, Pwc)
    check_argument(
        not (isinstance(x, Pwc) and isinstance(y, Stf))
        and not (isinstance(x, Stf) and isinstance(y, Pwc)),
        f"You can't apply the {op_name} operation between a Pwc and an Stf.",
        {"x": x, "y": y},
    )

    # operation(Stf, Stf)
    if isinstance(x, Stf) and isinstance(y, Stf):
        check_argument(
            name is None, "You can't assign a name to an Stf node.", {"name": name}
        )

        x_batch_shape, x_value_shape = validate_batch_and_value_shapes(x, "x")
        y_batch_shape, y_value_shape = validate_batch_and_value_shapes(y, "y")

        batch_shape, value_shape = validate_function_output_shapes(
            x_batch_shape,
            x_value_shape,
            y_batch_shape,
            y_value_shape,
            validate_value_shape=validate_value_shape,
        )
        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)

    # operation(Pwc, Pwc)
    if isinstance(x, Pwc) and isinstance(y, Pwc):
        check_argument(
            np.isclose(np.sum(x.durations), np.sum(y.durations)),
            "Both Pwc terms must have the same total duration.",
            {"x": x, "y": y},
        )

        x_batch_shape, x_value_shape = validate_batch_and_value_shapes(x, "x")
        y_batch_shape, y_value_shape = validate_batch_and_value_shapes(y, "y")

        batch_shape, value_shape = validate_function_output_shapes(
            x_batch_shape,
            x_value_shape,
            y_batch_shape,
            y_value_shape,
            validate_value_shape=validate_value_shape,
        )

        durations = mesh_pwc_durations([x, y])
        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )

    # operation(Stf, Tensor) or operation(Tensor, Stf)
    if isinstance(x, Stf) or isinstance(y, Stf):
        check_argument(
            name is None, "You can't assign a name to an Stf node.", {"name": name}
        )

        if isinstance(x, Stf):
            f_batch_shape, f_value_shape = validate_batch_and_value_shapes(x, "x")
            f_name = "x"
            check_numeric_numpy_array(y, "y")
            t_shape = validate_shape(y, "y")
            t_name = "y"
            tensor_first = False
        else:
            check_numeric_numpy_array(x, "x")
            t_shape = validate_shape(x, "x")
            t_name = "x"
            f_batch_shape, f_value_shape = validate_batch_and_value_shapes(y, "y")
            f_name = "y"
            tensor_first = True

        batch_shape, value_shape = validate_tensor_and_function_output_shapes(
            t_shape,
            f_batch_shape,
            f_value_shape,
            t_name,
            f_name,
            validate_value_shape=validate_value_shape,
            tensor_first=tensor_first,
        )

        return Stf(_operation, value_shape=value_shape, batch_shape=batch_shape)

    # operation(Pwc, Tensor) or operation(Tensor, Pwc)
    if isinstance(x, Pwc) or isinstance(y, Pwc):
        if isinstance(x, Pwc):
            f_batch_shape, f_value_shape = validate_batch_and_value_shapes(x, "x")
            f_name = "x"
            durations = x.durations
            check_numeric_numpy_array(y, "y")
            t_shape = validate_shape(y, "y")
            t_name = "y"
            tensor_first = False
        else:
            check_numeric_numpy_array(x, "x")
            t_shape = validate_shape(x, "x")
            t_name = "x"
            f_batch_shape, f_value_shape = validate_batch_and_value_shapes(y, "y")
            f_name = "y"
            durations = y.durations
            tensor_first = True

        batch_shape, value_shape = validate_tensor_and_function_output_shapes(
            t_shape,
            f_batch_shape,
            f_value_shape,
            t_name,
            f_name,
            validate_value_shape=validate_value_shape,
            tensor_first=tensor_first,
        )

        return Pwc(
            _operation,
            value_shape=value_shape,
            durations=durations,
            batch_shape=batch_shape,
        )

    # operation(Tensor, Tensor)
    check_numeric_numpy_array(x, "x")
    check_numeric_numpy_array(y, "y")
    x_shape = validate_shape(x, "x")
    y_shape = validate_shape(y, "y")
    shape = validate_value_shape(x_shape, y_shape, "x", "y")
    return Tensor(_operation, shape=shape)


class Addition(Node):
    """
    Calculate the element-wise sum between numbers, np.ndarrays, Tensors, Pwcs,
    or Stfs. You can also use the arithmetic operator ``+`` to calculate their sum.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The left summand, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The right summand, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise sum :math:`x+y`.
    """

    name = "add"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "add", **kwargs)


class Subtraction(Node):
    """
    Calculate the element-wise difference between numbers, np.ndarrays, Tensors, Pwcs,
    or Stfs. You can also use the arithmetic operator ``-`` to calculate their difference.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The minuend, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The subtrahend, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise difference :math:`x-y`.
    """

    name = "subtract"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "subtract", **kwargs)


class Multiplication(Node):
    r"""
    Calculate the element-wise product between numbers, np.ndarrays, Tensors,
    Pwcs, or Stfs. You can also use the arithmetic operator ``*`` to calculate their product.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The left factor, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The right factor, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise product :math:`x \times y`.
    """

    name = "multiply"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "multiply", **kwargs)


class TrueDivision(Node):
    """
    Calculate the element-wise division between numbers, np.ndarrays, Tensors, Pwcs,
    or Stfs. You can also use the arithmetic operator ``/`` to calculate their division.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The numerator, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The denominator, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise division :math:`x/y`.

    See Also
    --------
    Graph.floordiv : Divide two values and take the floor of the result.
    """

    name = "truediv"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "truediv", **kwargs)


class FloorDivision(Node):
    r"""
    Calculate the element-wise rounded-down division between real numbers, np.ndarrays, Tensors,
    Pwcs, or Stfs. You can also use the arithmetic operator ``//`` to calculate their floor
    division.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The numerator, :math:`x`. Must be real-valued.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The denominator, :math:`y`. Must be real-valued.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise rounded-down division :math:`\lfloor x/y \rfloor`.

    See Also
    --------
    Graph.truediv : Divide two values.
    """

    name = "floordiv"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "floordiv", **kwargs)


class Exponentiation(Node):
    """
    Calculate the element-wise power between numbers, np.ndarrays, Tensors, Pwcs,
    or Stfs. You can also use the arithmetic operator ``**`` to calculate their power.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The base, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The exponent, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise power :math:`x^y`.

    Warnings
    --------
    This function considers that the zeroth power of zero (:math:`0^0`) is
    undefined. This means that you might see an error if you attempt to
    fetch an object that contains :math:`0^0`.
    """

    name = "pow"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.ARITHMETIC_FUNCTIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "pow", **kwargs)


class ComplexValue(Node):
    """
    Create element-wise complex values from real numbers, np.ndarrays, Tensors,
    Pwcs, or Stfs, that is, the real and imaginary parts.

    Considering numbers and np.ndarrays as Tensors, if the two objects are of the same type,
    so is the returned object. If the objects have different types, Pwcs and Stfs can operate
    with a tensor (returning a Pwc or Stf, respectively).

    This operation supports broadcasting between the different objects. When operating a tensor-like
    object with an Stf or a Pwc, the time dimension of the latter is ignored.

    Parameters
    ----------
    x : number or np.ndarray or Tensor or Pwc or Stf
        The real part, :math:`x`.
    y : number or np.ndarray or Tensor or Pwc or Stf
        The imaginary part, :math:`y`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x` nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The element-wise complex number :math:`x+iy`.

    See Also
    --------
    Graph.abs : Absolute value of a complex object.
    Graph.angle : Argument of a complex object.
    Graph.conjugate : Conjugate of a complex object.
    Graph.imag : Imaginary part of a complex object.
    Graph.real : Real part of a complex object.
    """

    name = "complex_value"
    args = [
        forge.arg("x", type=NumericOrFunction),
        forge.arg("y", type=NumericOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.MATH]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return _create_flexible_binary_node_data(_operation, "complex", **kwargs)


class Matmul(Node):
    """
    Calculate the matrix multiplication between np.ndarrays, Tensors,
    Pwcs, or Stfs. You can also use the arithmetic operator ``@``
    to calculate their matrix multiplication.

    If any of the inputs is a Pwc or Stf, the output is also a
    Pwc or Stf (mixing Pwcs and Stfs is not supported).
    Otherwise, the output is a Tensor.

    This operation supports broadcasting between the batch dimensions of
    the two input objects. All the dimensions with the exception of the two
    innermost ones (where the matrix multiplication is performed) are
    considered batch dimensions.

    When operating a tensor-like object with an Stf or a Pwc, the time
    dimension of the latter is ignored.

    Parameters
    ----------
    x : np.ndarray or Tensor or Pwc or Stf
        The left multiplicand. It must be a matrix (or batch of matrices)
        and its last dimension must be the same as the second-to-last
        dimension of `y`.
    y : np.ndarray or Tensor or Pwc or Stf
        The right multiplicand. It must be a matrix (or batch of matrices)
        and its second-to-last dimension must be the same as the last
        dimension of `x`.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x`
        nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The matrix product of the input objects. If any of the input
        objects is a Pwc or Stf, the returned objects has the same
        type. Otherwise, it is a Tensor.

    See Also
    --------
    Graph.einsum : Tensor contraction via Einstein summation convention.
    """

    name = "matmul"
    args = [
        forge.arg("x", type=TensorLikeOrFunction),
        forge.arg("y", type=TensorLikeOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        def _validate_value_shape(x_shape, y_shape, x_name, y_name):
            if len(x_shape) < 2 or len(y_shape) < 2:
                raise QctrlException(
                    f"The shapes {x_shape} of {x_name} and {y_shape} of"
                    f" {y_name} must have at least two dimensions."
                )
            if get_broadcasted_shape(x_shape[:-2], y_shape[:-2]) is None:
                raise QctrlException(
                    f"The shapes {x_shape} of {x_name} and {y_shape} of"
                    f" {y_name} must be broadcastable (except in the last two"
                    " dimensions)."
                )

            if x_shape[-1] != y_shape[-2]:
                raise QctrlException(
                    f"The last dimension of the shape {x_shape} of {x_name} must be"
                    f" equal to the second-to-last dimension of the shape {y_shape}"
                    f" of {y_name}."
                )

            # Due to the previous checks, this should never fail.
            trailing_shape = validate_broadcasted_shape(
                x_shape[:-2], y_shape[:-2], f"{x_name} (batch)", f"{y_name} (batch)"
            )

            return trailing_shape + (x_shape[-2], y_shape[-1])

        return _create_flexible_binary_node_data(
            _operation,
            "matmul",
            validate_value_shape=_validate_value_shape,
            can_x_be_number=False,
            can_y_be_number=False,
            **kwargs,
        )


class Kron(Node):
    """
    Calculate the Kronecker product between np.ndarrays, Tensors,
    Pwcs, or Stfs.

    If any of the inputs is a Pwc or Stf, the output is also a
    Pwc or Stf (mixing Pwcs and Stfs is not supported).
    Otherwise, the output is a Tensor.

    This operation supports broadcasting between the batch dimensions of
    the two input objects. All the dimensions with the exception of the two
    innermost ones (where the Kronecker product is performed) are
    considered batch dimensions.

    When operating a tensor-like object with an Stf or a Pwc, the time
    dimension of the latter is ignored.

    Parameters
    ----------
    x : np.ndarray or Tensor or Pwc or Stf
        The left multiplicand. It must be a have at least two dimensions.
    y : np.ndarray or Tensor or Pwc or Stf
        The right multiplicand. It must be a have at least two dimensions.
    name : str or None, optional
        The name of the node. You can only provide a name if neither `x`
        nor `y` are Stfs.

    Returns
    -------
    Tensor or Pwc or Stf
        The Kronecker product of the input objects. If any of the input
        objects is a Pwc or Stf, the returned objects has the same
        type. Otherwise, it is a Tensor.

    See Also
    --------
    Graph.embed_operators : Embed operators into a larger Hilbert space.
    Graph.kronecker_product_list : Kronecker product of a list of operators.
    Graph.pauli_kronecker_product : Embed Pauli matrices into a larger Hilbert space.
    """

    name = "kron"
    args = [
        forge.arg("x", type=TensorLikeOrFunction),
        forge.arg("y", type=TensorLikeOrFunction),
    ]
    rtype = Union[Tensor, Pwc, Stf]
    categories = [Category.LINEAR_ALGEBRA]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        def _validate_value_shape(x_shape, y_shape, x_name, y_name):
            if len(x_shape) < 2 or len(y_shape) < 2:
                raise QctrlException(
                    f"The shapes {x_shape} of {x_name} and {y_shape} of"
                    f" {y_name} must have at least two dimensions."
                )
            if get_broadcasted_shape(x_shape[:-2], y_shape[:-2]) is None:
                raise QctrlException(
                    f"The shapes {x_shape} of {x_name} and {y_shape} of"
                    f" {y_name} must be broadcastable (except in the last two"
                    " dimensions)."
                )

            # Due to the previous check, this should never fail.
            trailing_shape = validate_broadcasted_shape(
                x_shape[:-2], y_shape[:-2], f"{x_name}", f"{y_name}"
            )

            return trailing_shape + (
                x_shape[-2] * y_shape[-2],
                x_shape[-1] * y_shape[-1],
            )

        return _create_flexible_binary_node_data(
            _operation,
            "kron",
            validate_value_shape=_validate_value_shape,
            can_x_be_number=False,
            can_y_be_number=False,
            **kwargs,
        )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

__version__ = "1.0.0"

from qctrlworkflowclient.utils import PackageInfo as _PackageInfo
from qctrlworkflowclient.utils import check_package_version as _check_package_version

from boulderopal import (
    closed_loop,
    cloud,
    gradient_free,
    graph,
    ions,
    noise_reconstruction,
    optimization,
    signals,
    stochastic,
    superconducting,
)
from boulderopal._utils import (
    cite,
    print_package_versions,
)
from boulderopal.gradient_free import run_gradient_free_optimization
from boulderopal.graph import (
    Graph,
    execute_graph,
)
from boulderopal.optimization import run_optimization
from boulderopal.stochastic import run_stochastic_optimization

_check_package_version(
    _PackageInfo(
        name="Boulder Opal client",
        install_name="boulder-opal",
        import_name="boulderopal",
        changelog_url="https://docs.q-ctrl.com/boulder-opal/changelog",
    )
)


__all__ = [
    "Graph",
    "cite",
    "closed_loop",
    "cloud",
    "execute_graph",
    "gradient_free",
    "graph",
    "ions",
    "noise_reconstruction",
    "optimization",
    "print_package_versions",
    "run_gradient_free_optimization",
    "run_optimization",
    "run_stochastic_optimization",
    "signals",
    "stochastic",
    "superconducting",
]



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

from qctrlcommons.preconditions import check_argument

from boulderopal._core import run_stochastic_optimization_workflow
from boulderopal._validation.basic import (
    ScalarDType,
    nullable,
    validate_enum,
)
from boulderopal._validation.graph import (
    check_cost_node,
    check_cost_node_for_optimization_graph,
    check_initial_value_for_optimization_node,
    check_optimization_node_in_graph,
    validate_output_node_names,
)
from boulderopal.optimization._optimization import HistoryScope
from boulderopal.stochastic._optimizer import Adam

if TYPE_CHECKING:
    from boulderopal.graph.graph import Graph


def run_stochastic_optimization(
    graph: "Graph",
    cost_node_name: str,
    output_node_names: str | list[str],
    optimizer: Optional[Adam | str] = None,
    iteration_count: int = 1000,
    target_cost: Optional[float] = None,
    cost_history_scope: HistoryScope = HistoryScope.NONE,
    seed: Optional[int] = None,
) -> dict:
    r"""
    Perform gradient-based stochastic optimization of generic real-valued functions.

    Use this function to determine a choice of variables that minimizes the
    value of a stochastic scalar real-valued cost function of those variables. You
    express that cost function as a graph describing how the input variables
    and stochastic variables are transformed into the cost value.

    Parameters
    ----------
    graph : Graph
        The graph describing the cost :math:`C(\mathbf v, \boldsymbol \beta)` and outputs
        :math:`\{F_j(\mathbf v)\}` as functions of the optimizable input variables
        :math:`\mathbf v`. The graph must contain nodes with names :math:`s` (giving the
        cost function) and :math:`\{s_j\}` (giving the output functions).
    cost_node_name : str
        The name :math:`s` of the real-valued scalar graph node that defines the cost
        function :math:`C(\mathbf v, \boldsymbol \beta)` to be minimized.
    output_node_names : str or list[str]
        The names :math:`\{s_j\}` of the graph nodes that define the output functions
        :math:`\{F_j(\mathbf v)\}`. The function evaluates these using the optimized
        variables and returns them in the output. If any of the output nodes depend
        on random nodes, the random values used to calculate the output might not
        correspond to the values used to calculate the final cost.
    optimizer : Adam or str or None
        The optimizer used for the stochastic optimization. You can either create an Adam optimizer
        for this parameter, or pass the string from a previous optimization result that represents
        state of the optimizer (the value with the key `state` in the returned dictionary from
        this function) to restart the optimization.
        Note that you can only resume a previous optimization if you pass the same graph.
        Defaults to a new Adam optimizer.
    iteration_count : int
        The number :math:`N` of iterations the optimizer performs until it halts.
        The function returns the results from the best iteration (the one with
        the lowest cost). Defaults to 1000.
    target_cost : float or None, Optional
        A target value of the cost that you can set as an early stop condition for the
        optimizer. If the cost becomes equal or smaller than this value, the
        optimization halts. Defaults to None, which means that this function runs
        until the `iteration_count` is reached.
    cost_history_scope : HistoryScope, Optional
        Configuration for the scope of the returned cost history data.
        Use this to select how you want the history data to be returned.
        Defaults to no cost history data returned.
    seed : int or None, optional
        Seed for the random number generator used in the optimizer.
        If set, must be a non-negative integer.
        Use this option to generate deterministic results from the optimizer.
        Note that if your graph contains nodes generating random samples,
        you need to also set seeds for those nodes to ensure a reproducible result.

    Returns
    -------
    dict
        A dictionary containing the optimization result, with the following keys:

        ``cost``
            The minimum cost function value :math:`C(\mathbf v_\mathrm{optimized})`
            achieved across all optimization iterations.
        ``output``
            The dictionary giving the value of each requested output node, evaluated
            at the optimized variables, namely :math:`\{s_j: F_j(\mathbf v_\mathrm{optimized})\}`.
            The keys of the dictionary are the names :math:`\{s_j\}` of the output nodes.
            If any of the output nodes depend on random nodes, the random values used to
            calculate the output might not correspond to the values used to calculate the best cost.
        ``state``
            The encoded optimizer state. You can use this parameter to resume the optimization from
            the current step.
        ``cost_history``
            The evolution of the cost function across all optimization iterations.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    :func:`boulderopal.stochastic.Adam`
        Create an Adam optimizer for stochastic optimization.
    :func:`boulderopal.closed_loop.optimize`
        Run a closed-loop optimization to find a minimum of the given cost function.
    :func:`boulderopal.closed_loop.step`
        Perform a single step in a closed-loop optimization.
    :func:`boulderopal.execute_graph`
        Evaluate generic functions.
    :func:`boulderopal.run_gradient_free_optimization`
        Perform model-based optimization without using gradient values.
    :func:`boulderopal.run_optimization`
        Perform gradient-based deterministic optimization of generic real-valued functions.

    Notes
    -----

    Given a cost function :math:`C(\mathbf v, \boldsymbol \beta)` of optimization
    variables :math:`\mathbf v` and stochastic variables :math:`\boldsymbol \beta`,
    this function computes an estimate :math:`\mathbf v_\mathrm{optimized}` of
    :math:`\mathrm{argmin}_{\mathbf v} C(\mathbf v, \boldsymbol \beta)`, namely the
    choice of variables :math:`\mathbf v` that minimizes :math:`C(\mathbf v, \boldsymbol \beta)`
    with noise through the stochastic variables :math:`\boldsymbol \beta`.
    The function then calculates the values of arbitrary output functions
    :math:`\{F_j(\mathbf v_\mathrm{optimized})\}` with that choice of variables.

    This function represents the cost and output functions as nodes of a graph.
    This graph defines the input variables :math:`\mathbf v` and stochastic variables
    :math:`\boldsymbol \beta`, and how these variables are transformed into the corresponding
    cost and output quantities. You build the graph from primitive nodes defined in the
    Graph object. Each such node, which can be identified by a name, represents a function
    of the previous nodes in the graph (and thus, transitively, a function of the input variables).
    You can use any named scalar real-valued node :math:`s` as the cost function, and any named
    nodes :math:`\{s_j\}` as outputs.

    After you provide a cost function :math:`C(\mathbf v, \boldsymbol \beta)` (via a graph),
    this function runs the optimization process for :math:`N` iterations, each with random
    stochastic variables, to identify local minima of the stochastic cost function, and then takes
    the variables corresponding to the best such minimum as :math:`\mathbf v_\mathrm{optimized}`.

    Note that this function only performs a single optimization run. That means, if you provide
    lists of initial values for optimization variables in the graph, only the first one for
    each variable will be used.

    A common use-case for this function is to determine controls for a quantum system that yield
    an optimal gate subject to noise: the variables :math:`\mathbf v` parameterize the controls to
    be optimized, and the cost function :math:`C(\mathbf v, \boldsymbol \beta)` is the operational
    infidelity describing the quality of the resulting gate relative to a target gate with noise
    through the stochastic variables :math:`\boldsymbol \beta`. When combined with the node
    definitions in the Boulder Opal package, which make it convenient to define such cost functions,
    this function provides a highly configurable framework for quantum control that encapsulates
    other common tools such as batch gradient ascent pulse engineering [1]_.

    References
    ----------
    .. [1] `R. Wu, H. Ding, D. Dong, and X. Wang, Physical Review A 99, 042327 (2019).
            <https://doi.org/10.1103/PhysRevA.99.042327>`_

    Examples
    --------
    Perform a simple stochastic optimization.

    >>> graph = bo.Graph()
    >>> x = graph.optimization_variable(1, -1, 1, name="x")
    >>> cost = (x - 0.5) ** 2
    >>> cost.name = "cost"
    >>> result = bo.run_stochastic_optimization(
    ...     graph=graph, cost_node_name="cost", output_node_names="x"
    ... )
    >>> result["cost"], result["output"]
        (0.0, {'x': {'value': array([0.5])}})

    To have a better understanding of the optimization landscape, you can use the
    `cost_history_scope` parameter to retrieve the cost history information from the optimizer.
    See the reference for the available options.
    For example, to retrieve all available history information:

    >>> history_result = bo.run_stochastic_optimization(
    ...     graph=graph,
    ...     cost_node_name="cost",
    ...     output_node_names="x",
    ...     cost_history_scope="ALL",
    ... )

    You can then access the history information from the `cost_history` key.
    We only show here the last two records to avoid a lengthy output.

    >>> history_result["cost_history"]["iteration_values"][-2:]
        [1.9721522630525295e-31, 1.9721522630525295e-31]
    >>> history_result["cost_history"]["historical_best"][-2:]
        [0.0, 0.0]

    See also the `How to optimize controls robust to strong noise sources
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-
    controls-robust-to-strong-noise-sources>`_ user guide.
    """

    check_optimization_node_in_graph(graph)
    check_initial_value_for_optimization_node(graph)
    check_cost_node(cost_node_name, graph)

    output_node_names = validate_output_node_names(output_node_names, graph)
    check_cost_node_for_optimization_graph(graph, cost_node_name, output_node_names)

    if optimizer is None:
        optimizer = Adam()
    else:
        check_argument(
            isinstance(optimizer, (Adam, str)),
            "The optimizer must either be an Adam optimizer or"
            "a string to represent the state of the optimizer.",
            {"optimizer": optimizer},
        )
    assert optimizer is not None
    if isinstance(optimizer, Adam):
        optimizer_configuration: dict[str, Any] = {
            "adam": {"learning_rate": optimizer.learning_rate}
        }
    else:
        optimizer_configuration = {"state": optimizer}

    return run_stochastic_optimization_workflow(
        graph=graph,
        optimizer=optimizer_configuration,
        cost_node_name=cost_node_name,
        output_node_names=output_node_names,
        iteration_count=ScalarDType.INT(iteration_count, "iteration_count", min_=0),
        target_cost=nullable(ScalarDType.REAL, target_cost, "target_cost"),
        cost_history_scope=validate_enum(HistoryScope, cost_history_scope),
        seed=nullable(ScalarDType.INT, seed, "seed", min_=0, min_inclusive=True),
    )
